# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""YubiKey helpers.
A layer of helpers between ATBU and yubikey-manager.
"""

from typing import Any, Tuple
from atbu.common.exception import exc_to_string, InvalidFunctionArgument

from .exception import (
    YubiKeyBackendNotAvailableError,
    YubiKeyNotPressedTimeout,
)

_IS_YUBIKEY_REQUIRED = False
_IS_YUBIKEY_INFRA_INIT = False


def set_require_yubikey(is_required: bool):
    global _IS_YUBIKEY_REQUIRED
    _IS_YUBIKEY_REQUIRED = is_required


def is_yubikey_required():
    return _IS_YUBIKEY_REQUIRED


def setup_yubikey_infra():
    # pylint: disable=import-outside-toplevel,unused-import
    global _IS_YUBIKEY_INFRA_INIT
    if _IS_YUBIKEY_INFRA_INIT:
        return
    try:
        global ykman
        global yubikit
        import ykman
        import ykman.device
        import yubikit.core
        import yubikit.yubiotp

        _IS_YUBIKEY_INFRA_INIT = True
    except Exception as ex:
        raise YubiKeyBackendNotAvailableError(
            f"Failure access YubiKey backend. "
            f"Ensure you have yubikey-manager installed: "
            f"pip install yubikey-manager. {exc_to_string(ex)}"
        ).with_traceback(ex.__traceback__) from ex


def _get_list_yubikey_devices() -> list[Tuple[Any, Any]]:
    # pylint: disable=undefined-variable
    setup_yubikey_infra()
    l = ykman.device.list_all_devices()
    return l if l else None


def is_a_yubikey_present():
    l = _get_list_yubikey_devices()
    if l is None:
        return False
    return len(l) > 0


def get_first_yubikey_device():
    l = _get_list_yubikey_devices()
    return l[0] if l else (None, None)


def get_max_challenge_size():
    # pylint: disable=undefined-variable
    return yubikit.yubiotp.HMAC_CHALLENGE_SIZE


def challenge_response(
    challenge: bytes,
    slot_num: int = 2,
):
    # pylint: disable=undefined-variable
    setup_yubikey_infra()
    if len(challenge) > get_max_challenge_size():
        raise InvalidFunctionArgument(
            f"The challenge must be {get_max_challenge_size()} bytes or less."
        )
    device, _ = get_first_yubikey_device()
    if device is None:
        raise YubiKeyBackendNotAvailableError("First YubiKey not found.")
    with device.open_connection(yubikit.core.otp.OtpConnection) as c:
        s = yubikit.yubiotp.YubiOtpSession(c)
        try:
            response = s.calculate_hmac_sha1(
                slot=slot_num,  # Will use CONFIG_SLOT.CHAL_HMAC_n
                challenge=challenge,
            )
            return response
        except yubikit.core.TimeoutError as ex:
            raise YubiKeyNotPressedTimeout(
                f"The YubiKey was not pressed in time."
            ).with_traceback(ex.__traceback__) from ex



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""List various ATBU items: storage definitions, specific backups, files within.
"""
import logging
from tabulate import tabulate

from atbu.common.exception import InvalidStateError
from .constants import CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX
from .backup_selections import (
    get_storage_defintions_from_sbs_list_list,
    user_specifiers_to_selections,
)
from .backup_core import StorageDefinition


def handle_list(args):
    logging.debug(f"handle_list")

    specifiers: list[str] = args.specifiers
    if not specifiers:
        specifiers = [f"{CONFIG_SECTION_STORAGE_DEFINITION_SPECIFIER_PREFIX}:*"]

    sbs_list_list = user_specifiers_to_selections(
        specifiers=specifiers, no_selections_ok=True
    )

    headers = [
        "Storage Definition",
        "Provider",
        "Container",
        "Interface",
        "Encrypted",
        "Persisted IV",
    ]
    records = []
    storage_def_list: list[
        StorageDefinition
    ] = get_storage_defintions_from_sbs_list_list(sbs_list_list=sbs_list_list)
    for storage_def in storage_def_list:
        storage_def_name = storage_def.storage_def_name
        records.append(
            [
                storage_def_name,
                storage_def.driver_factory.provider_name,
                storage_def.container_name,
                storage_def.driver_factory.interface_type,
                storage_def.is_encryption_used,
                storage_def.storage_persisted_encryption_IV,
            ]
        )
    logging.info(
        "\n" + tabulate(tabular_data=records, headers=headers, tablefmt="simple")
    )

    for sbs_list in sbs_list_list:  # pylint: disable=not-an-iterable
        storage_def_name = None
        for sbi in sbs_list:
            if sbi.specific_backup_info is None:
                continue
            if not storage_def_name:
                storage_def_name = sbi.storage_def_name
                logging.info(
                    f"Specific backups from storage definition '{storage_def_name}'"
                )
            elif storage_def_name != sbi.storage_def_name:
                raise InvalidStateError(
                    f"Expected entire list to belong to the same storage def '{storage_def_name}'"
                )
            logging.info(f"  {sbi.specific_backup_name}")
            list_fi_sorted = sorted(sbi.selected_fi.values(), key=lambda fi: fi.path)
            for sb_fi in list_fi_sorted:
                logging.info(f"    {sb_fi.path}")



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""E2E tests for persistent file information.
"""

# pylint: disable=unnecessary-pass
# pylint: disable=unused-import,wrong-import-position
# pylint: disable=line-too-long

from asyncio import InvalidStateError
from collections import namedtuple
from dataclasses import dataclass
from io import SEEK_END, SEEK_SET
import os
from pathlib import Path
from random import randint
import random
import re
import string
from typing import Any
import logging

LOGGER = logging.getLogger(__name__)
import pytest
from pytest import (
    LogCaptureFixture,
    CaptureFixture,
    fail,
    raises,
    Pytester,
    FixtureRequest,
    Config,
    RunResult,
    ExitCode,
)

from atbu.tools.backup.global_hasher import GlobalHasherDefinitions
from atbu.tools.backup.constants import (
    ATBU_PERSIST_TYPE_PER_BOTH,
    ATBU_PERSIST_TYPE_PER_DIR,
    ATBU_PERSIST_TYPE_PER_FILE,
)
from atbu.tools.persisted_info.file_info import (
    ATBU_PERSISTENT_INFO_EXTENSION,
    FileInformation,
    FileInformationPersistent,
)
from atbu.tools.persisted_info.database import (
    FileInformationDatabaseCollection,
)

from .common_helpers import (
    DirInfo,
    StaticTestValues,
    create_test_data_directory_default_levels,
    duplicate_tree,
    establish_random_seed,
    get_rel_path,
    get_rel_path_nc,
    run_atbu,
)

SIZE_1MB = 1024 * 1024
SIZE_2MB = 2 * SIZE_1MB
SIZE_BINARY_CONTENTS_1 = SIZE_1MB
SIZE_BINARY_CONTENTS_2 = SIZE_1MB + 1
SIZE_BINARY_CONTENTS_3 = SIZE_1MB + 2
SIZE_BINARY_CONTENTS_4 = SIZE_2MB
SIZE_BINARY_CONTENTS_5 = SIZE_2MB + 1
BYTES_16 = b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
TEXT_FILE_CONTENTS_1 = "What a wonderful day!"
TEXT_FILE_CONTENTS_2 = "What a wonderful day!!"
TEXT_FILE_CONTENTS_4 = "What a wonderful day!!!"
TEXT_FILE_CONTENTS_5 = "What a wonderful evening!"

STATE_ORIGINAL = "original"
STATE_DELETED = "deleted"
STATE_BITROT = "bitrot"


def setup_module(module):  # pylint: disable=unused-argument
    pass


def teardown_module(module):  # pylint: disable=unused-argument
    pass


LayoutEntry = namedtuple("LayoutEntry", "directory, file_name, content_type, content")

basic_dir_layout1: list[LayoutEntry] = [
    LayoutEntry(
        "Folder1",
        "Folder1File1-SIZE_BINARY_CONTENTS_1.bin",
        "binary",
        SIZE_BINARY_CONTENTS_1,
    ),
    LayoutEntry(
        "Folder1", "Folder1File2-TEXT_FILE_CONTENTS_1.txt", "text", TEXT_FILE_CONTENTS_1
    ),
    LayoutEntry(
        "Folder1", "Folder1File3-TEXT_FILE_CONTENTS_4.txt", "text", TEXT_FILE_CONTENTS_4
    ),
    LayoutEntry(
        "Folder1",
        "Folder1File4-SIZE_BINARY_CONTENTS_5.bin",
        "binary",
        SIZE_BINARY_CONTENTS_5,
    ),
    LayoutEntry(
        "Folder2",
        "Folder2File1-SIZE_BINARY_CONTENTS_2.bin",
        "binary",
        SIZE_BINARY_CONTENTS_2,
    ),
    LayoutEntry(
        "Folder2",
        "Folder2File2-SIZE_BINARY_CONTENTS_3.bin",
        "binary",
        SIZE_BINARY_CONTENTS_3,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File1-SIZE_BINARY_CONTENTS_1.bin",
        "binary",
        SIZE_BINARY_CONTENTS_1,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File2-SIZE_BINARY_CONTENTS_4.bin",
        "binary",
        SIZE_BINARY_CONTENTS_4,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File3-SIZE_BINARY_CONTENTS_4.bin",
        "binary",
        SIZE_BINARY_CONTENTS_4,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File4-SIZE_BINARY_CONTENTS_5.bin",
        "binary",
        SIZE_BINARY_CONTENTS_5,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File5-TEXT_FILE_CONTENTS_4.txt",
        "text",
        TEXT_FILE_CONTENTS_4,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File6-TEXT_FILE_CONTENTS_1.txt",
        "text",
        TEXT_FILE_CONTENTS_1,
    ),
    LayoutEntry(
        "Folder2/Folder2Folder1",
        "Folder2Folder1File7-TEXT_FILE_CONTENTS_2.txt",
        "text",
        TEXT_FILE_CONTENTS_2,
    ),
]

OutputExtractionDefinition = namedtuple(
    "OutputExtractionDefinition", "name, regex, cls"
)
UpdatedFileInfo = namedtuple("UpdatedFileHash", "path, hash_name, hash")
AddedFileInfo = namedtuple("AddedFileInfo", "path, hash_name, hash")
UpToDateFileInfo = namedtuple("UpToDateFileInfo", "path, hash_name, hash")
AddedFileInfoDetailed = namedtuple(
    "AddedFileInfo", "path, config_path, size_in_bytes, last_modified, hash_name, hash"
)
output_extraction_definitions: list[OutputExtractionDefinition] = [
    OutputExtractionDefinition(
        name="UpdatedFileInfo",
        # Example: Updated info: path=<path_name_here> sha256=<digest_here>
        regex=re.compile(
            f".*The.*file info was updated: path=([^\\s]+)\\s+([^=]+)=([0-9a-fA-F]+).*"
        ),
        cls=UpdatedFileInfo,
    ),
    OutputExtractionDefinition(
        name="AddedFileInfo",
        # Example: Updated info: path=<path_name_here> sha256=<digest_here>
        regex=re.compile(
            f".*The.*file info was added: path=([^\\s]+)\\s+([^=]+)=([0-9a-fA-F]+).*"
        ),
        cls=AddedFileInfo,
    ),
    OutputExtractionDefinition(
        name="UpToDateFileInfo",
        # Example: Updated info: path=<path_name_here> sha256=<digest_here>
        regex=re.compile(
            f".*The.*file info was up to date: path=([^\\s]+)\\s+([^=]+)=([0-9a-fA-F]+).*"
        ),
        cls=UpToDateFileInfo,
    ),
    OutputExtractionDefinition(
        name="AddedFileInfoDetailed",
        # Example: Adding file information to results: path=<path_name_here>\nconfig_path=<path_name_here>\ninfo_current...\n  sizeinbytes=1048576\n  lastmodified=2022/03/28-00:17:40\n  sha256=<digest_here>\ninfo_history...\n  INFO.0000:\n    sizeinbytes=1048576\n    lastmodified=2022/03/28-00:17:40\n    sha256=<digest_here>'
        regex=re.compile(
            f".*The.*file info was added:"
            f".*path=([^\\s]+).*\n.*config_path=([^\\s]+).*\n.*info_current.*\n.*sizeinbytes=(\\d+).*\n.*lastmodified=([^\\s]+).*\n.*\\s([^\\s=]+)=([0-9a-zA-Z]+).*\n.*info_history.*"
        ),
        cls=AddedFileInfoDetailed,
    ),
]


def get_persist_type_prefix(persist_types: list[str]) -> str:
    if len(persist_types) == 1:
        persist_type_prefix = persist_types[0]
    else:
        persist_type_prefix = ATBU_PERSIST_TYPE_PER_BOTH
    return persist_type_prefix


@dataclass
class SpecificLayoutEntry:
    parent: object
    dir_path: Path
    file_path: Path
    config_file_path: Path
    file_name: Path
    content_type: string
    content: Any
    state: string
    old_digest: string
    validated: bool = False


class SpecificLayout:
    def __init__(
        self,
        root_path: Path,
        layout: list[LayoutEntry],
        persist_types: list[str] = None,
    ):
        if persist_types is None:
            persist_types = [ATBU_PERSIST_TYPE_PER_FILE]
        self.persist_types = persist_types
        self.root_path = root_path
        self.dbc: FileInformationDatabaseCollection = None
        self.dict_ncpath_to_fi: dict[str, FileInformationPersistent] = None
        self._entries: list[SpecificLayoutEntry] = []
        for le in layout:
            dir_path: Path = root_path / le.directory
            file_path: Path = dir_path / le.file_name
            config_file_path: Path = file_path.with_suffix(
                file_path.suffix + ATBU_PERSISTENT_INFO_EXTENSION
            )
            self._entries.append(
                SpecificLayoutEntry(
                    parent=self,
                    dir_path=dir_path,
                    file_path=file_path,
                    config_file_path=config_file_path,
                    file_name=le.file_name,
                    content_type=le.content_type,
                    content=le.content,
                    state=STATE_ORIGINAL,
                    old_digest=None,
                    validated=False,
                )
            )

    def load_db(self):
        if self.dbc is not None:
            return self.dbc
        self.dbc = FileInformationDatabaseCollection(
            source_path=self.root_path, persist_types=self.persist_types
        )
        self.dbc.load()
        self.dict_ncpath_to_fi = self.dbc.get_dict_nc_path_to_fi()
        return self.dbc

    def reload_db(self):
        self.dbc = None
        self.load_db()

    def get_file_info_from_db_cfg(self, sle: SpecificLayoutEntry):
        self.load_db()
        nc_file_path = os.path.normcase(str(sle.file_path))
        fi = self.dict_ncpath_to_fi.get(nc_file_path)
        return fi

    def get_file_info(self, sle: SpecificLayoutEntry, is_read: bool = False):
        if ATBU_PERSIST_TYPE_PER_FILE in self.persist_types:
            fi = FileInformationPersistent(path=str(sle.file_path))
            if is_read:
                fi.read_info_data_file()
        elif ATBU_PERSIST_TYPE_PER_DIR in self.persist_types:
            fi = self.get_file_info_from_db_cfg(sle=sle)
            if fi is None:
                fi = FileInformationPersistent(path=str(sle.file_path))
        return fi

    def is_config_present(self, sle: SpecificLayoutEntry):
        for pt in self.persist_types:
            if pt == ATBU_PERSIST_TYPE_PER_DIR:
                return self.get_file_info_from_db_cfg(sle=sle) is not None
            elif pt == ATBU_PERSIST_TYPE_PER_FILE:
                return sle.config_file_path.exists()
            else:
                raise InvalidStateError(f"Expected value persist type but got {pt}")

    def delete_config_file(self, sle: SpecificLayoutEntry):
        for pt in self.persist_types:
            if pt == ATBU_PERSIST_TYPE_PER_DIR:
                # Do nothing for DB case because, in
                # real user scenario, DB does not
                # auto-update, user must run update-digests
                # which tests within do themselves.
                pass
            elif pt == ATBU_PERSIST_TYPE_PER_FILE:
                # Simulate user deleted config file (along
                # with the data file).
                sle.config_file_path.unlink()
            else:
                raise InvalidStateError(f"Expected value persist type but got {pt}")

    @property
    def is_per_file_config(self) -> bool:
        return ATBU_PERSIST_TYPE_PER_FILE in self.persist_types

    @property
    def entries(self):
        return self._entries

    def __iter__(self):
        return self.entries.__iter__()

    def __getitem__(self, key):
        return self.entries.__getitem__(key)


def reset_specific_layout_validated_state(layout: SpecificLayout):
    for sle in layout:
        sle.validated = False


def create_binary_file(file_path: Path, size):
    LOGGER.debug(f"Creating binary file: path={file_path} size={size}")
    with open(file_path, "wb") as f:
        b = BYTES_16 * int(SIZE_1MB / len(BYTES_16))  # 1MB
        while size >= SIZE_1MB:
            f.write(b)
            size -= SIZE_1MB
        while size >= len(BYTES_16):
            f.write(BYTES_16)
            size -= len(BYTES_16)
        while size > 0:
            size_to_write = min(size, len(BYTES_16))
            f.write(BYTES_16[:size_to_write])
            size -= size_to_write


def create_text_file(file_path: Path, contents):
    LOGGER.debug(
        f"Creating text file: path={file_path} size={len(contents)} contents={contents[:128]}"
    )
    with open(file_path, "wt", encoding="utf-8") as f:
        f.write(contents)


def create_test_dir_layout(root_path: Path, specific_layout: SpecificLayout):
    LOGGER.debug(f"Creating test dir layout in {root_path}")
    for sle in specific_layout:
        LOGGER.debug(f"Creating directory: {sle.dir_path}")
        sle.dir_path.mkdir(exist_ok=True)
        if sle.content_type == "binary":
            create_binary_file(sle.file_path, sle.content)
        elif sle.content_type == "text":
            create_text_file(sle.file_path, sle.content)


def extract_info_from_output(output_lines: list[str]):
    info = []
    for output in output_lines:
        for oed in output_extraction_definitions:
            mo = oed.regex.match(output)
            if mo is not None:
                info.append(oed.cls(*mo.groups()))
    return info


def is_digest_in_output(digest: str, output_lines: list[str]):
    digest = digest.lower()
    for output in output_lines:
        if output.lower().find(digest) != -1:
            return True
    return False


def verify_specific_layout_validated(specific_layout: SpecificLayout):
    for sle in specific_layout:
        assert sle.validated


def verify_output_file_hash_info(
    output_info_list: list,
    file_path: Path,
    state: str,
    digest_from_config_file: str,
    hashing_algo_name: str,
):
    for o in output_info_list:
        if isinstance(o, UpdatedFileInfo):
            # if isinstance(oi, UpdatedFileInfo):
            ufh: UpdatedFileInfo = o
            if Path(ufh.path) == file_path:
                assert state != STATE_DELETED
                assert ufh.hash_name == hashing_algo_name
                assert ufh.hash == digest_from_config_file
                return True
        elif isinstance(o, AddedFileInfo):
            afi: AddedFileInfo = o
            if Path(afi.path) == file_path:
                assert state != STATE_DELETED
                assert afi.hash_name == hashing_algo_name
                assert afi.hash == digest_from_config_file
                return True
        elif isinstance(o, UpToDateFileInfo):
            afi: UpToDateFileInfo = o
            if Path(afi.path) == file_path:
                assert state != STATE_DELETED
                assert afi.hash_name == hashing_algo_name
                assert afi.hash == digest_from_config_file
                return True
    if state != STATE_DELETED:
        fail(
            f"File name and hash not found in test output and not expected to be {state}: path={file_path}"
        )


def verify_expected_vs_actual(specific_layout: SpecificLayout, info: list):
    primary_hashing_algo_name = (
        GlobalHasherDefinitions().get_primary_hashing_algo_name()
    )
    for sle in specific_layout:
        sl: SpecificLayout = sle.parent
        fidf = None
        digest = None
        calculated_hash = None
        if sle.state == STATE_ORIGINAL:
            assert sle.file_path.exists()
            assert sl.is_config_present(sle)
            fidf = sl.get_file_info(sle=sle, is_read=True)
            assert str(sle.file_path) == fidf.path
            if sl.is_per_file_config:
                assert sle.config_file_path == Path(fidf.info_data_file_path)
                assert str(sle.file_path) == fidf.path
                assert sle.config_file_path == Path(fidf.info_data_file_path)
            digest = fidf.get_current_digest()
            verify_output_file_hash_info(
                output_info_list=info,
                file_path=sle.file_path,
                state=sle.state,
                digest_from_config_file=digest,
                hashing_algo_name=primary_hashing_algo_name,
            )
            calculated_hash = FileInformation(path=str(sle.file_path)).primary_digest
            assert digest == calculated_hash
            sle.validated = True
        elif sle.state == STATE_DELETED:
            assert not sle.file_path.exists()
            assert not sl.is_config_present(sle)
            fidf = sl.get_file_info(sle=sle)
            with raises(FileNotFoundError):
                fidf.read_info_data_file()
            verify_output_file_hash_info(
                output_info_list=info,
                file_path=sle.file_path,
                state=sle.state,
                digest_from_config_file="",
                hashing_algo_name=primary_hashing_algo_name,
            )
            sle.validated = True
        elif sle.state == STATE_BITROT:
            assert sle.file_path.exists()
            assert sl.is_config_present(sle)
            fidf = sl.get_file_info(sle=sle, is_read=True)
            digest = fidf.get_current_digest()
            verify_output_file_hash_info(
                output_info_list=info,
                file_path=sle.file_path,
                state=sle.state,
                digest_from_config_file=digest,
                hashing_algo_name=primary_hashing_algo_name,
            )
            calculated_hash = FileInformation(path=str(sle.file_path)).primary_digest
            assert digest == calculated_hash
            assert (
                sle.old_digest != digest
            )  # pre-bitrot digest does not equal post-bitrot digest.
            sle.validated = True
        else:
            fail("Unknown state.")
    verify_specific_layout_validated(specific_layout=specific_layout)


def verify_config_files_exist(root_path: Path, layout: list[LayoutEntry]):
    specific_layout = SpecificLayout(root_path=root_path, layout=layout)
    for sle in specific_layout:
        assert sle.file_path.exists()
        assert sle.config_file_path.exists()


def verify_config_files_exist2(
    persist_types: list[str],
    root_path: Path,
    layout: list[LayoutEntry],
):
    specific_layout = SpecificLayout(
        root_path=root_path,
        layout=layout,
        persist_types=persist_types,
    )
    for sle in specific_layout:
        assert sle.file_path.exists()
        assert specific_layout.is_config_present(sle)


def create_layout1(
    persist_types: list[str],
    tmp_path: Path,
) -> tuple[Path, SpecificLayout, Path, SpecificLayout,]:
    locA_path = tmp_path / "LocationA"
    locB_path = tmp_path / "LocationB"
    locA_path.mkdir()
    locB_path.mkdir()
    locA_specific_layout = SpecificLayout(
        root_path=locA_path,
        layout=basic_dir_layout1,
        persist_types=persist_types,
    )
    locB_specific_layout = SpecificLayout(
        root_path=locB_path,
        layout=basic_dir_layout1,
        persist_types=persist_types,
    )
    create_test_dir_layout(root_path=locA_path, specific_layout=locA_specific_layout)
    create_test_dir_layout(root_path=locB_path, specific_layout=locB_specific_layout)
    return locA_path, locA_specific_layout, locB_path, locB_specific_layout


def create_layout1_with_config_files(
    persist_types: list[str], tmp_path: Path, pytester: Pytester
):
    persist_type_prefix = get_persist_type_prefix(persist_types=persist_types)
    locA_path, locA_specific_layout, locB_path, locB_specific_layout = create_layout1(
        persist_types=persist_types,
        tmp_path=tmp_path,
    )
    argv = [
        "update-digests",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="update-digests",
    )
    assert rr.ret == ExitCode.OK
    verify_config_files_exist2(
        persist_types=persist_types, root_path=locA_path, layout=basic_dir_layout1
    )
    verify_config_files_exist2(
        persist_types=persist_types, root_path=locB_path, layout=basic_dir_layout1
    )
    return locA_path, locA_specific_layout, locB_path, locB_specific_layout


def update_digests(
    persist_types: list[str],
    location: str,
    is_digest_change_detection: bool,
    tmp_path: Path,
    pytester: Pytester,
):
    persist_type_prefix = get_persist_type_prefix(persist_types=persist_types)

    if is_digest_change_detection:
        # The default is date/time and size changes (faster), but
        # at caller's discretion, use digest-based (recalc all digests).
        argv = [
            "update-digests",
            "--change-detection-type",
            "digest",
            f"{persist_type_prefix}:",
            location,
        ]
    else:
        argv = [
            "update-digests",
            f"{persist_type_prefix}:",
            location,
        ]

    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="update-digests",
    )
    assert rr.ret == ExitCode.OK


def update_digests_specific_layout(
    specific_layout: SpecificLayout,
    is_digest_change_detection: bool,
    update_only_db_cfg: bool,
    tmp_path: Path,
    pytester: Pytester,
):
    if not update_only_db_cfg or not specific_layout.is_per_file_config:
        # DB case will not have its config "deleted" until rescan.
        # Rescan to update digests/db.
        update_digests(
            persist_types=specific_layout.persist_types,
            location=str(specific_layout.root_path),
            is_digest_change_detection=is_digest_change_detection,
            tmp_path=tmp_path,
            pytester=pytester,
        )
        if not specific_layout.is_per_file_config:
            specific_layout.reload_db()


persist_type_parameters = [
    pytest.param(
        [ATBU_PERSIST_TYPE_PER_DIR],
        id=ATBU_PERSIST_TYPE_PER_DIR.replace("-", "_"),
    ),
    pytest.param(
        [ATBU_PERSIST_TYPE_PER_FILE],
        id=ATBU_PERSIST_TYPE_PER_FILE.replace("-", "_"),
    ),
]


@pytest.mark.parametrize(
    "persist_types",
    persist_type_parameters,
)
def test_create_info(
    persist_types: list[str],
    tmp_path: Path,
    pytester: Pytester,
):
    create_layout1_with_config_files(persist_types, tmp_path, pytester)


@pytest.mark.parametrize(
    "persist_types",
    persist_type_parameters,
)
def test_diff(
    persist_types: list[str],
    tmp_path: Path,
    capsys: CaptureFixture,  # pylint: disable=unused-argument
    caplog: LogCaptureFixture,
    pytester: Pytester,
):
    caplog.set_level("DEBUG")
    (
        locA_path,
        locA_specific_layout,
        locB_path,
        locB_specific_layout,
    ) = create_layout1_with_config_files(
        persist_types=persist_types,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    persist_type_prefix = get_persist_type_prefix(persist_types=persist_types)

    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff1",
    )
    assert rr.ret == ExitCode.OK
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    reset_specific_layout_validated_state(layout=locA_specific_layout)
    reset_specific_layout_validated_state(layout=locB_specific_layout)

    caplog.clear()
    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff2",
    )
    assert rr.ret == ExitCode.OK
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    pass


@pytest.mark.parametrize(
    "persist_types",
    persist_type_parameters,
)
def test_diff_with_deleted_files(
    persist_types: list[str],
    tmp_path: Path,
    capsys: CaptureFixture,  # pylint: disable=unused-argument
    caplog: LogCaptureFixture,
    pytester: Pytester,
):
    persist_type_prefix = get_persist_type_prefix(persist_types=persist_types)

    (
        locA_path,
        locA_specific_layout,
        locB_path,
        locB_specific_layout,
    ) = create_layout1_with_config_files(
        persist_types=persist_types, tmp_path=tmp_path, pytester=pytester
    )

    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff1",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    # Delete from Location A

    reset_specific_layout_validated_state(layout=locA_specific_layout)
    reset_specific_layout_validated_state(layout=locB_specific_layout)

    sle: SpecificLayoutEntry = locA_specific_layout[3]
    sle.file_path.unlink()
    locA_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    sle: SpecificLayoutEntry = locA_specific_layout[6]
    sle.file_path.unlink()
    locA_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    update_digests_specific_layout(
        specific_layout=locA_specific_layout,
        is_digest_change_detection=False,
        update_only_db_cfg=True,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    caplog.clear()
    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff2",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    # Delete from Location B

    reset_specific_layout_validated_state(layout=locA_specific_layout)
    reset_specific_layout_validated_state(layout=locB_specific_layout)

    sle: SpecificLayoutEntry = locB_specific_layout[3]
    sle.file_path.unlink()
    locB_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    sle: SpecificLayoutEntry = locB_specific_layout[7]
    sle.file_path.unlink()
    locB_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    update_digests_specific_layout(
        specific_layout=locB_specific_layout,
        is_digest_change_detection=False,
        update_only_db_cfg=True,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    caplog.clear()
    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff3",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    # Delete from Location A and B

    reset_specific_layout_validated_state(layout=locA_specific_layout)
    reset_specific_layout_validated_state(layout=locB_specific_layout)

    sle: SpecificLayoutEntry = locA_specific_layout[4]
    sle.file_path.unlink()
    locA_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    sle: SpecificLayoutEntry = locB_specific_layout[8]
    sle.file_path.unlink()
    locB_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    update_digests_specific_layout(
        specific_layout=locA_specific_layout,
        is_digest_change_detection=False,
        update_only_db_cfg=True,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    update_digests_specific_layout(
        specific_layout=locB_specific_layout,
        is_digest_change_detection=False,
        update_only_db_cfg=True,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    caplog.clear()
    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff4",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    # Delete remaining files of SIZE_BINARY_CONTENTS_5
    # No digests of SIZE_BINARY_CONTENTS_5 should exists afterwards.
    # layout entries at index 3 and 9 are for content #5.

    reset_specific_layout_validated_state(layout=locA_specific_layout)
    reset_specific_layout_validated_state(layout=locB_specific_layout)

    sle: SpecificLayoutEntry = locA_specific_layout[9]
    file_info = locA_specific_layout.get_file_info(sle, is_read=True)
    the_digest = file_info.get_current_digest()
    assert is_digest_in_output(the_digest, rr.outlines)
    sle.file_path.unlink()
    locA_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    sle: SpecificLayoutEntry = locB_specific_layout[9]
    file_info = locB_specific_layout.get_file_info(sle, is_read=True)
    assert the_digest == file_info.get_current_digest()
    sle.file_path.unlink()
    locB_specific_layout.delete_config_file(sle)
    sle.state = STATE_DELETED

    update_digests_specific_layout(
        specific_layout=locA_specific_layout,
        is_digest_change_detection=False,
        update_only_db_cfg=True,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    update_digests_specific_layout(
        specific_layout=locB_specific_layout,
        is_digest_change_detection=False,
        update_only_db_cfg=True,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    caplog.clear()
    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff5",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    assert not is_digest_in_output(the_digest, rr.outlines)
    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)
    pass


@pytest.mark.parametrize(
    "persist_types",
    persist_type_parameters,
)
def test_diff_bit_rot(
    persist_types: list[str],
    tmp_path: Path,
    capsys: CaptureFixture,  # pylint: disable=unused-argument
    caplog: LogCaptureFixture,
    pytester: Pytester,
):
    persist_type_prefix = get_persist_type_prefix(persist_types=persist_types)

    (
        locA_path,
        locA_specific_layout,
        locB_path,
        locB_specific_layout,
    ) = create_layout1_with_config_files(
        persist_types=persist_types, tmp_path=tmp_path, pytester=pytester
    )

    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff1",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)

    # Simulate bitrot in Location B

    reset_specific_layout_validated_state(layout=locA_specific_layout)
    reset_specific_layout_validated_state(layout=locB_specific_layout)

    sle: SpecificLayoutEntry = locB_specific_layout[3]
    sle.old_digest = locB_specific_layout.get_file_info(sle).primary_digest
    sr: os.stat_result = os.stat(sle.file_path)
    with open(sle.file_path, "r+b") as the_file:
        the_file.seek(0, SEEK_END)
        file_size = the_file.tell()
        mod_pos = int(file_size / 2)
        the_file.seek(mod_pos, SEEK_SET)
        b = the_file.read(1)
        the_file.seek(mod_pos, SEEK_SET)
        the_file.write(bytes([b[0] + 1]))
    os.utime(
        path=sle.file_path,
        times=(
            sr.st_atime,
            sr.st_mtime,
        ),
    )
    sle.state = STATE_BITROT

    update_digests_specific_layout(
        specific_layout=locB_specific_layout,
        is_digest_change_detection=True,
        update_only_db_cfg=False,
        tmp_path=tmp_path,
        pytester=pytester,
    )

    caplog.clear()
    argv = [
        "diff",
        f"{persist_type_prefix}:",
        str(locA_path),
        str(locB_path),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="diff2-bitrot-detect",
    )
    info = extract_info_from_output(output_lines=rr.outlines)

    verify_expected_vs_actual(specific_layout=locA_specific_layout, info=info)
    verify_expected_vs_actual(specific_layout=locB_specific_layout, info=info)
    pass


@pytest.mark.parametrize(
    "persist_types",
    persist_type_parameters,
)
def test_arrange_basic(
    persist_types: list[str],
    tmp_path: Path,
    caplog: LogCaptureFixture,
    pytester: Pytester,
):
    establish_random_seed(tmp_path)  # bytes([0,1,2,3])

    persist_type_prefix = get_persist_type_prefix(persist_types=persist_types)
    is_per_file = True if persist_type_prefix == ATBU_PERSIST_TYPE_PER_FILE else False

    original_test_data_directory = tmp_path / "OriginalTestDataDir"
    current_template_root = tmp_path / "TemplateRoot"
    outdated_target_source_root = tmp_path / "TargetSourceRoot"
    arranged_target_dest_root = tmp_path / "TargetDestRoot"
    undofile_path = tmp_path / "UndoFiles"
    undofile_path.mkdir(parents=True, exist_ok=False)

    initial_dir_size_defs = [
        StaticTestValues(values=list(range(64)), some_limit=2),
    ]

    dirs_list, files_list = create_test_data_directory_default_levels(
        path_to_dir=original_test_data_directory,
        file_size_defs=initial_dir_size_defs,
    )

    duplicate_tree(
        src_dir=original_test_data_directory,
        dst_dir=current_template_root,
    )

    duplicate_tree(
        src_dir=current_template_root,
        dst_dir=outdated_target_source_root,
    )

    with DirInfo(dir_path=current_template_root) as template_di:
        template_di.gather_info()

        removed_or_moved_old_rel_paths = set()
        removed_or_moved_new_rel_paths = set()

        # Delete a file.
        files_deleted, files_remaining = template_di.delete_randomly_chosen_files(
            num_to_delete=1,
        )
        deleted_file_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=files_deleted[0],
            )
        )
        removed_or_moved_old_rel_paths.add(deleted_file_rel_path)

        idx_available_for_tests = [*range(0, len(template_di.file_list))]
        random.shuffle(idx_available_for_tests)

        # Move a file to a newly created directory.
        idx_to_move = idx_available_for_tests.pop()
        move_src_path = template_di.file_list[idx_to_move].path
        move_src_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=move_src_path,
            )
        )
        removed_or_moved_old_rel_paths.add(move_src_rel_path)
        basename_to_move = os.path.basename(move_src_path)
        dest_dir = os.path.join(current_template_root, "NewDir")
        move_dest_path = os.path.join(dest_dir, basename_to_move)
        move_dest_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=move_dest_path,
            )
        )
        removed_or_moved_new_rel_paths.add(move_dest_rel_path)
        os.renames(
            old=move_src_path,
            new=move_dest_path,
        )

        # Rename an existing file, keeping in the same directory.
        idx_to_rename = idx_available_for_tests.pop()
        rename_src_path = template_di.file_list[idx_to_rename].path
        rename_src_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=rename_src_path,
            )
        )
        removed_or_moved_old_rel_paths.add(rename_src_rel_path)
        base, ext = os.path.splitext(os.path.basename(rename_src_path))
        new_basename_of_rename = f"Rename-{base}-Rename{ext}"
        rename_dest_path = os.path.join(os.path.dirname(rename_src_path), new_basename_of_rename)
        rename_dest_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=rename_dest_path,
            )
        )
        removed_or_moved_new_rel_paths.add(rename_dest_rel_path)
        os.renames(
            old=rename_src_path,
            new=rename_dest_path,
        )

        # Rename an existing file, moving it to another directory.
        idx_to_move_rename = idx_available_for_tests.pop()
        move_rename_src_path = template_di.file_list[idx_to_move_rename].path
        move_rename_src_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=move_rename_src_path,
            )
        )
        removed_or_moved_old_rel_paths.add(move_rename_src_rel_path)
        base, ext = os.path.splitext(os.path.basename(move_rename_src_path))
        new_basename_of_move_rename = f"Rename-{base}-Rename{ext}"
        move_rename_dest_dir = os.path.join(current_template_root, "NewDirRenamedFile")
        move_rename_dest_path = os.path.join(move_rename_dest_dir, new_basename_of_move_rename)
        move_rename_dest_rel_path = os.path.normcase(get_rel_path(
                root_path=current_template_root,
                path_within_root=move_rename_dest_path,
            )
        )
        removed_or_moved_new_rel_paths.add(move_rename_dest_rel_path)
        os.renames(
            old=move_rename_src_path,
            new=move_rename_dest_path,
        )

    add_files_size_defs = [
        StaticTestValues(values=list(range(1000,1020)), some_limit=2),
    ]

    dirs_added, files_added = create_test_data_directory_default_levels(
        path_to_dir=current_template_root,
        file_size_defs=add_files_size_defs,
        add_files_to_existing=True
    )

    files_remaining.extend(files_added)

    caplog.clear()
    argv = [
        "arrange",
        f"{persist_type_prefix}:",
        str(current_template_root),
        str(outdated_target_source_root),
        str(arranged_target_dest_root),
        "--undofile",
        str(undofile_path / "arrange1_undo.json"),
        "--loglevel",
        "DEBUG",
    ]
    rr = run_atbu(
        pytester,
        tmp_path,
        *argv,
        log_base_name="arrange1",
    )
    assert rr.ret == ExitCode.OK

    with (
        DirInfo(dir_path=original_test_data_directory) as original_di,
        DirInfo(dir_path=current_template_root) as template_di,
        DirInfo(dir_path=outdated_target_source_root) as target_source_di,
        DirInfo(dir_path=arranged_target_dest_root) as target_dest_di,
    ):
        original_di.gather_info(start_gathering_digests=True)
        template_di.gather_info()
        target_source_di.gather_info()
        target_dest_di.gather_info(start_gathering_digests=True)

        original_set = original_di.get_nc_rel_path_set()
        template_set = template_di.get_nc_rel_path_set()
        target_source_set = target_source_di.get_nc_rel_path_set()
        target_dest_set = target_dest_di.get_nc_rel_path_set()

        #
        # Sanity and test checks for deleted/moved files.
        #

        # The deleted file is in original, not in updated template and target destination.
        assert deleted_file_rel_path in original_set
        assert deleted_file_rel_path in target_source_set
        assert deleted_file_rel_path not in template_set
        assert deleted_file_rel_path not in target_dest_set

        # The original moved file location is in original, not in template and target dest. 
        assert move_src_rel_path in original_set
        assert move_src_rel_path not in template_set
        if is_per_file:
            assert f"{move_src_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" not in template_set
        assert move_src_rel_path not in target_source_set
        if is_per_file:
            assert f"{move_src_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" not in target_source_set
        assert move_src_rel_path not in target_dest_set
        if is_per_file:
            assert f"{move_src_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" not in target_dest_set

        # The moved file dest is not in the original but in template and target dest.
        assert move_dest_rel_path not in original_set
        assert move_dest_rel_path in template_set
        if is_per_file:
            assert f"{move_dest_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" in template_set
        assert move_dest_rel_path not in target_source_set
        if is_per_file:
            assert f"{move_dest_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" not in target_source_set
        assert move_dest_rel_path in target_dest_set
        if is_per_file:
            assert f"{move_dest_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" in target_dest_set

        # The target destination less the template should contain 0 paths because
        # the destination is built based on what is observed in the template.
        target_dest_less_template = target_dest_set.difference(template_set)
        assert len(target_dest_less_template) == 0

        # The original less the newly created target destination set should contain any paths
        # which were removed or moved.
        original_less_target_dest = original_set.difference(target_dest_set)
        assert original_less_target_dest == removed_or_moved_old_rel_paths
        #removed_or_moved_old_rel_paths !!! remove
        #removed_or_moved_new_rel_paths

        # Files added to the template did not exist in the target source, which means they
        # will not exist in the target destination. If 'per-file' is used, .atbu files will
        # exist for all added files which doubles the count (i.e., if 20 files added, 40 
        # additional files will exist in the template). Verify all files added are present
        # in the delta with the target destination. For per-dir, the .atbudb will be present
        # hence +1.
        dest_files_added_count = len(files_added)*2 if is_per_file else len(files_added) + 1
        template_less_target_dest = template_set.difference(target_dest_set)
        assert len(template_less_target_dest) == dest_files_added_count
        files_added_nc_rel_path_set = set([get_rel_path_nc(current_template_root, fa) for fa in files_added])
        assert files_added_nc_rel_path_set.issubset(template_less_target_dest)


        target_source_less_target_dest = target_source_set.difference(target_dest_set)
        assert len(target_source_less_target_dest) == 2
        assert deleted_file_rel_path in target_source_less_target_dest
        if is_per_file:
            assert f"{deleted_file_rel_path}{ATBU_PERSISTENT_INFO_EXTENSION}" in target_source_less_target_dest

        original_rp_fi_dict = original_di.get_nc_rel_path_dict()
        for dest_rp, dest_fi in target_dest_di.get_nc_rel_path_dict().items():
            orig_fi = original_rp_fi_dict.get(dest_rp)
            if orig_fi is None:
                if dest_rp == move_dest_rel_path:
                    orig_fi = original_rp_fi_dict.get(move_src_rel_path)
                elif dest_rp == rename_dest_rel_path:
                    orig_fi = original_rp_fi_dict.get(rename_src_rel_path)
                elif dest_rp == move_rename_dest_rel_path:
                    orig_fi = original_rp_fi_dict.get(move_rename_src_rel_path)
            if orig_fi is not None:
                assert dest_fi == orig_fi
                if is_per_file:
                    fip = FileInformationPersistent(path=dest_fi.path)
                    assert fip.info_data_file_exists()
                    fip.read_info_data_file()
                    assert dest_fi.digest == fip.primary_digest
            elif os.path.splitext(dest_rp)[1] == ATBU_PERSISTENT_INFO_EXTENSION:
                continue
            else:
                fail(f"Expected to validate all target destination files: {dest_rp}")
        pass  # pylint: disable=unnecessary-pass



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Main entry point, argument parsing.
"""
# pylint: disable=line-too-long

import argparse
import logging
import multiprocessing
import os
import re
import sys
from typing import Union

from atbu.common.exception import (
    exc_to_string,
    AtbuException,
    InvalidCommandLineArgument,
    InvalidStateError,
    QueueListenerNotStarted,
    SingletonAlreadyCreated,
)
from atbu.common.profile import EasyProfile
from atbu.mp_pipeline.mp_global import (
    deinitialize_logging,
    get_verbosity_level,
    global_init,
    initialize_logging_basic,
    initialize_logging,
    remove_created_logging_handlers,
    remove_root_stream_handlers,
)
from atbu.common.hasher import (
    DEFAULT_HASH_ALGORITHM,
)
from atbu.tools.backup.config import (
    set_automated_testing_mode,
    get_automated_testing_mode,
    register_storage_def_config_override,
)


from .constants import *
from .exception import CredentialSecretFileNotFoundError, YubiKeyBackendNotAvailableError
from .global_hasher import GlobalHasherDefinitions
from .backup_core import (
    BACKUP_COMPRESSION_CHOICES,
    BACKUP_COMPRESSION_DEFAULT,
    BACKUP_INFO_EXTENSION,
)
from .backup_cmdline import handle_backup
from .restore import handle_restore, handle_decrypt
from .verify import handle_verify
from .recover import handle_recover
from .list_items import handle_list
from .yubikey_helpers import set_require_yubikey, setup_yubikey_infra
from .creds_cmdline import handle_creds
from ..persisted_info.file_info import (
    CHANGE_DETECTION_CHOICES,
    CHANGE_DETECTION_TYPE_DATESIZE,
    CHANGE_DETECTION_TYPE_DIGEST,
    CHANGE_DETECTION_TYPE_FORCE,
)
from ..persisted_info.update_digests import handle_update_digests
from ..persisted_info.database import handle_savedb
from ..persisted_info.diff import handle_diff, DIFF_COMMAND_CHOICES
from ..persisted_info.arrange import handle_arrange

class AtbuRawTextHelpFormatter(argparse.RawTextHelpFormatter):
    """This class modifies RawTextHelpFormatter help text to insert
    blank lines with a heading before each group of ATBU argparse subparsers.

    Instead of something like this...
        ...
        backup              Backup files to a local file system folder or the cloud.
        restore             Restore selected files from a backup.
        ...

    ...a heading is inserted with blank lines as follows...

        ...
                        Backup/Restore/Verify sub-commands
                        -----------------------------------

    backup              Backup files to a local file system folder or the cloud.
    restore             Restore selected files from a backup.
        ...
    """
    def format_help(self):

        help_text = super().format_help()

        m = re.search(
            pattern=r"( +backup +)Backup files to a local file system folder or the cloud.",
            string=help_text,
        )
        if m is not None:
            help_padding = m.end(1) - m.start(1)
            help_text = (
                help_text[:m.start()] +
                "\n" +
                " " * help_padding + "Backup/Restore/Verify sub-commands\n" +
                " " * help_padding + "-----------------------------------\n" +
                "\n" +
                help_text[m.start():]
            )

        m = re.search(
            pattern=r"( +update-digests +)Update persistent file information in the specified directories.",
            string=help_text,
        )
        if m is not None:
            help_padding = m.end(1) - m.start(1)
            help_text = (
                help_text[:m.start()] +
                "\n" +
                " " * help_padding + "Persistent file information-related sub-commands (unrelated to backup/restore)\n" +
                " " * help_padding + "------------------------------------------------------------------------------\n" +
                "\n" +
                help_text[m.start():]
            )

        m = re.search(
            pattern=r"( +help +)Show help on specific subjects:",
            string=help_text,
        )
        if m is not None:
            help_padding = m.end(1) - m.start(1)
            help_text = (
                help_text[:m.start()] +
                "\n" +
                " " * help_padding + "Other sub-commands\n" +
                " " * help_padding + "------------------\n" +
                "\n" +
                help_text[m.start():]
            )

        return help_text


#
# Common help strings used by argparse configuration further below.
#
storage_def_specifier_help = """
<storage-def-specifier>:
    <storage-def-specifier> is at least a storage specifier, possibly followed by backup and file selectors.
    For the case of a backup, it is simply a storage specifier. For the case of restore, verify and some list
    commands, it is usually a storage specifier, a backup selector, one or more file selectors, all of which
    determine what storage files to restore, verify, list, etc.
        <storage-def-specifier> is shorthand for the following...
        <storage-specifier> [backup:<backup_specifier> [files:<file_specifier> [files:<file_specifier>] ... ] ] ...
    Not all commands allow all portions of a <storage-def-specifier>. See command-specific help for clarification.
    Details:
        <storage_specifier>: storage-def:<storage-definition-name>
            where
                * <storage-definition-name> is either a folder of a local file system backup or
                  a the name you gave to a cloud storage configuration.
                * 'storage-def' can be abbreviated to 'storage' if desired.
            Examples:
                - storage-def:my_cloud_cfg
                - storage:my_cloud_cfg (same as storage-def:my_cloud_cfg)
                - d:\\MyLocalBackup
        backup:<backup_specifier>
            where
                * <backup_specifier> can be 'last' to indicate the most recent backup.
                * <backup_specifier> can be a glob-like selector such as '*20220509-142129*'.
            Examples:
                - backup:last is the most recent backup.
                - backup:MyBackup-20220509-142129 is a specific backup of MyBackup from May 9, 2022 at 14:21:29.
                - backup:MyBackup-2022* are all backups to MyBackup in the year 2022.
        files:<file_specifier>
            where
                * <file_specifier> is essentially a glob-like pattern indicating files to select.
            Examples:
                - files:* will select all files.
                - files:*.txt will select all files ending with ".txt"
"""

backup_information_switch_help = f"""
Backup Information (BI)
=======================
Backup information (BI) is the information that {ATBU_PROGRAM_NAME} stores during and at the end of a
backup. It is contained in *{BACKUP_INFO_EXTENSION} files which are stored in BI directories. This
section discusses BI, BI-related files and locations.

By default, {ATBU_PROGRAM_NAME} stores BI in the following locations:

    For cloud storage definitions:
        Primary: <HOME>/{ATBU_DEFAULT_CONFIG_DIR_NAME}/{ATBU_DEFAULT_BACKUP_INFO_SUBDIR}
        Secondary: None

    For file system storage definitions:
        Primary: <HOME>/{ATBU_DEFAULT_CONFIG_DIR_NAME}/{ATBU_DEFAULT_BACKUP_INFO_SUBDIR}
        Secondary: <backup-dest>/{ATBU_DEFAULT_CONFIG_DIR_NAME}/{ATBU_DEFAULT_BACKUP_INFO_SUBDIR}
        (Basically, file system storage backups store a copy on the destination storage media.)

You can override the defaults by specifying configuration settings in your backup's storage definition
configuration .json file. You can specify backup information directory configuration information in
the .json configuration's {CONFIG_SECTION_GENERAL} section. For example, consider the following "{CONFIG_VALUE_NAME_BACKUP_INFO_DIR}"
setting:

    ...
    "{CONFIG_SECTION_GENERAL}": {{
        "{CONFIG_VALUE_NAME_BACKUP_INFO_DIR}": [
            "F:\\MyBackupInfo"
        ]
    }}
    ...

If the above were added to a cloud storage definition configuration, {ATBU_PROGRAM_NAME} would end up using
the following backup information directories...

    For cloud storage definitions:
        Primary: <HOME>/{ATBU_DEFAULT_CONFIG_DIR_NAME}/{ATBU_DEFAULT_BACKUP_INFO_SUBDIR}
        Secondary: F:\\MyBackupInfo

Notice that the default primary is still the same but now there is a secondary when before there was
not. With a secondary backup information location in place, during backup {ATBU_PROGRAM_NAME} save the
primary copy to the primary location, and it will then copy that to the secondary location of
F:\\MyBackupInfo.

You can instruct {ATBU_PROGRAM_NAME} to not use any default backup information directories by specifying
the {CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS} setting. For example:

    ...
    "{CONFIG_SECTION_GENERAL}": {{
        "{CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS}": true,
        "{CONFIG_VALUE_NAME_BACKUP_INFO_DIR}": [
            "F:\\\\MyBackupInfo"
        ]
    }}
    ...

With the above {CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS}=true setting in place, {ATBU_PROGRAM_NAME} will use the following
backup information directory:

    Primary: F:\\MyBackupInfo

Notice the HOME directory location is no longer utilized. The only backup information directory is
F:\\MyBackupInfo which is now also the primary backup information directory.

Using Offline Config and BI
===========================
You can store a backup storage definition configuration and its secrets, along with all backup
information, offline on a removable drive, where the drive is only inserted into the client during
the time when backup operations are taking place. The following outlines how this can be accomplished.

You can use the {ATBU_PROGRAM_NAME} --config-file switch to reference a backup storage definition configuration
and its credentials as stored in a .json file. The .json file can be created by exporting your
backup storage definition using the "{ATBU_PROGRAM_NAME} creds export..." command.

After exporting to .json, you can then use "{ATBU_PROGRAM_NAME} <command_and_options> --config-file <json_file>"
to perform a backup/restore or other operations, where {ATBU_PROGRAM_NAME} will use the .json file's contents
to access the configuration, including its credentials/secrets. If the .json file was created from a
password-protected storage configuration, you will likewise be prompted to enter the password when
using --config-file to reference the .json backup storage definition configuration ("{ATBU_PROGRAM_NAME} creds export..."
exports password-protected credentials.).

By using an exported configuration, you can keep a configuration and its secrets off a client system,
and only make them available when performing backup operations. For example:

    1. atbu creds export storage-def:my-cloud-backup F:\\B\\my-cloud-backup-config.json, where F:
       is a USB Drive.

    2. Insert the USB Drive and perform a backup, referencing the F: .json config file.
       Example:
           atbu backup C:\\MyFiles storage-def:my-cloud-backup --full --config-file F:\\B\\my-cloud-backup-config.json

When you run the backup command in step 2, {ATBU_PROGRAM_NAME} will use the backup storage definition configuration
information, including secrets, stored in the "F:\\B\\my-cloud-backup-config.json" file.

Even though the above keeps the credentials stored offline, the backup information is still stored
in the default locations (i.e., the backup information is still stored on the local drive). To keep
the backup information offline (along with the storage definition configuration) you can perform the
following general steps:

    1. atbu creds export storage-def:my-cloud-backup F:\\B\\my-cloud-backup-config.json.

    2. Edit F:\\B\\my-cloud-backup-config.json, adding the following settings:
        ...
        \"{CONFIG_SECTION_GENERAL}\": {{
            \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS}\": true,
            \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR}\": [
                "{{CONFIG_DIR}}/atbu-backup-info"
            ]
        }}
        ...

    3. Perform a backup.
       Example:
           atbu backup C:\\MyFiles storage-def:my-cloud-backup --full --config-file F:\\B\\my-cloud-backup-config.json

The above 3 steps are similar to the prior example except that a new step #2 has been inserted. This
new step #2 modifies the exported .json configuration to disable default backup information
directories, while also specifying the primary (and only) backup information directory should be
"{{CONFIG_DIR}}/atbu-backup-info".

The "{{CONFIG_DIR}}" string is replaced with the directory where the configuration file itself is
located, F:\\B\\my-cloud-backup-config.json. Given this, the location of the backup information
directory would be "F:\\B\\atbu-backup-info\\...".

IMPORTANT: For large backups, ensure the location of any backup information directory is well-performing.
If you use a very slow USB Drive, reading/writing large backup information files can be time consuming.

In addition to "{{CONFIG_DIR}}", you can also specify "{{DEFAULT_CONFIG_DIR}}" which is the default
HOME {ATBU_DEFAULT_CONFIG_DIR_NAME} configuration directory.

Here is a summary of the replacements:

    "{{DEFAULT_CONFIG_DIR}}": The <HOME>/{ATBU_DEFAULT_CONFIG_DIR_NAME} location. 
    "{{CONFIG_DIR}}": The <config>/{ATBU_DEFAULT_CONFIG_DIR_NAME} location.

If you want to add multiple backup information directories to your configuration, simply add an
element to the .json array. For example:

        ...
        \"{CONFIG_SECTION_GENERAL}\": {{
            \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS}\": true,
            \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR}\": [
                "{{CONFIG_DIR}}/atbu-backup-info",
                "G:\\\\B\\\\atbu-backup-info"
            ]
        }}
        ...

Because \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS}\" is set to true, there are no defaults implicitly inserted,
so the primary backup information directory is therefore the first one explicitly configured, which
in this case is "{{CONFIG_DIR}}/atbu-backup-info". Additionally, there is one secondary backup
information directory specified. All together, the BI directory configuration ends up being as
follows:

    Primary: "{{CONFIG_DIR}}/atbu-backup-info"
    Secondary #1: "G:\\B\\atbu-backup-info"

Consider the following configuration that does not disable default backup information directories:
        ...
        \"{CONFIG_SECTION_GENERAL}\": {{
            \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR_NO_DEFAULTS}\": false,
            \"{CONFIG_VALUE_NAME_BACKUP_INFO_DIR}\": [
                "{{CONFIG_DIR}}/atbu-backup-info",
                "G:\\\\B\\\\atbu-backup-info"
            ]
        }}
        ...

The result would be the following:

    Primary: "{{DEFAULT_CONFIG_DIR}}/atbu-backup-info"
    Secondary #1: "{{CONFIG_DIR}}/atbu-backup-info"
    Secondary #2: "G:\\B\\atbu-backup-info"

The above result is because, with defaults active, the primary is always implicitly the default HOME
{ATBU_DEFAULT_CONFIG_DIR_NAME} configuration directory. This means the two directories explicitly specified become secondary
backup information directories.

Additional Information
======================
The {ATBU_PROGRAM_NAME} command uses local copies of *{BACKUP_INFO_EXTENSION} to understand the history of what is been backed
up. There must always be a "primary" backup information directory which contains the latest backup
information. The other BI directories than the primary are secondary and merely backup copies in
case the primary copies are unavailable. Having backup copies can allow for easier recovery or
browsing in certain cases, where it is desired to mitigate the need to use "{ATBU_PROGRAM_NAME} recover..." for
a full recovery. Recovery is generally used when all local copies of the latest backup information
have been lost or destroyed.

VERY IMPORTANT: Do not arbitrarily change a storage definition's backup information directory
configuration. Care must be taken to ensure, if you ever change a configuration's primary backup
information directory, you ensure that directory is populated with the latest copies of backup
information *{BACKUP_INFO_EXTENSION} files. You would do this by copying to the new location, or perhaps using
the "{ATBU_PROGRAM_NAME} recover..." command. Generally, once you establish backup information directories,
you should stick with the current settings unless changes are truly reasonable.

Regardless of any BI directory configuration options discussed in this help, {ATBU_PROGRAM_NAME} always backs
up *{BACKUP_INFO_EXTENSION} files at the end of each backup. These backup copies of *{BACKUP_INFO_EXTENSION} files are not
stored directly as *{BACKUP_INFO_EXTENSION}, but are instead encrypted (if encryption is enabled) and
stored as backup files for use primarily by the "{ATBU_PROGRAM_NAME} recover..." command during disaster
recovery. These backup copies of the *{BACKUP_INFO_EXTENSION} files are unrelated to the copies stored locally
at the client which is the primary subject of the above documentation.
"""

extra_help_subjects = {
    "specifiers": storage_def_specifier_help,
    "backup-info": backup_information_switch_help,
}

console_formatter = logging.Formatter("%(asctime)-15s %(threadName)s %(message)s")


def create_argparse():
    #
    # Root parser
    #
    parser = argparse.ArgumentParser(
        description=f"{ATBU_ACRONUM_U} v{ATBU_VERSION_STRING}",
        formatter_class=AtbuRawTextHelpFormatter,
    )

    # Uncomment to allow --debug-server (for use with VS Code pydebug)
    # Activate the debug server to listen on specified port, wait for a client connect.
    # parser.add_argument(
    #     "--debug-server",
    #     help=argparse.SUPPRESS,
    #     type=int,
    #     required=False,
    # )

    # Specified by automated tests to be used by this utility as needed for E2E tests.
    parser.add_argument(
        "--automated-testing",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )

    #
    # Common to all parser
    #
    parser_common = argparse.ArgumentParser(add_help=False)
    parser_common.add_argument(
        "--logfile",
        help="The location of log file path. If not specified, do not log to file.",
    )
    parser_common.add_argument(
        "--loglevel", help="level for logging (DEBUG, INFO, WARNING, ERROR, CRITICAL)."
    )
    parser_common.add_argument(
        "--log-console-detail",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""When logging to console use detailed format.
""",
    )
    parser_common.add_argument(
        "-v",
        "--verbosity",
        action="count",
        help="""increase verbosity with each usage (i.e., -vv is more verbose than -v).
""",
    )
    parser_common.add_argument(
        "--yk",
        action="store_true",
        default=False,
        help="""Require YubiKey HMAC-SHA1 challenge/response to access the backup encryption key.
When this option is specified, a YubiKey with slot 2 configured for HMAC-SHA1
challenge/response will be used/required along with your password to unlock a
backup encryption key. This means, for a password-protected backup, you will be
required to have your password and a YubiKey. This option is also used when
setting up a backup for use with Yubikey. You must have a YubiKey and you must
ensure yubikey-manager is installed (i.e., pip install yubikey-manager). See
documentation for further details.
""",
    )
    parser_common.add_argument(
        "--profile",
        action="store_true",
        default=False,
        help="""Run the command with an instance of cProfile.Profile() active, capturing the
results to file when the command completes.
""",
    )
    parser_common.add_argument(
        "--profile-stats",
        action="store_true",
        default=False,
        help="""When --profile is specified, display profile stats when the command completes.
""",
    )
    parser_common.add_argument(
        "--profile-file",
        help="""When specified, write the profile stats to the specified file name. The filename
must not already exist.
""",
    )

    #############################################################################################
    #                                backup-related argparse setup                              #
    #############################################################################################

    #
    # Common credential filename argument (currently used for import/export of credentials).
    #
    parser_credential_filename_positional = argparse.ArgumentParser(add_help=False)
    parser_credential_filename_positional.add_argument(
        "filename",
        help="The path to the credential file.",
    )

    #
    # An optional explicit configuration file which is searched for credentials.
    #
    parser_credential_filename_optional = argparse.ArgumentParser(add_help=False)
    parser_credential_filename_optional.add_argument(
        "--config-file", "-c",
        help=f"""The path to the credential configuration file. When specified, this config file is
the first "credential store" that is searched. To use this, use "{ATBU_PROGRAM_NAME} creds export..."
to export a backup storage definition configuration to a .json file. You can then
specify that json file as the argument to --config-file when you run backup, restore,
etc. commands.

When running a backup command with --config-file specified, ATBU will consider
credentials in that file for the storage definition in question. You can delete
the credentials from your local system's keystore and thereby keep your storage
config and credentials offline when not backing up.

When performing "{ATBU_PROGRAM_NAME} creds export...", if your secrets are password-protected,
the password-protection status will persist in the exported .json file. When you
use --config-file with a .json containing a password-protected backup config, you
will still be prompted for your password. This means if you keep the config on
removeable media which is lost, the credentials require a password to unlock.
IMPORTANT: Do not use a credential .json on a USB Drive as your main exported
config/secrets backup. Keep other copies in safe places!
""",
    )

    #
    # Common destination file overwrite argument.
    #
    parser_file_ovewrite = argparse.ArgumentParser(add_help=False)
    parser_file_ovewrite.add_argument(
        "--overwrite",
        help="Overwrite the destination file if it already exists.",
        action=argparse.BooleanOptionalAction,
        default=False,
    )

    # Add a subparser to act as a heading with blank lines before/after.
    subparsers = parser.add_subparsers(
        help=f"""""",
    )

    #
    # 'backup' subparser.
    #
    parser_backup = subparsers.add_parser(
        "backup",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Backup files to a local file system folder or the cloud.",
        parents=[parser_common, parser_credential_filename_optional],
    )
    group_backup_operation = parser_backup.add_mutually_exclusive_group(required=True)
    group_backup_operation.add_argument(
        f"-{ATBU_BACKUP_TYPE_FULL[0]}",
        f"--{ATBU_BACKUP_TYPE_FULL}",
        help="""Perform a full backup. All files will be backed up regardless of changes.
""",
        action="store_true",
    )
    group_backup_operation.add_argument(
        f"-{ATBU_BACKUP_TYPE_INCREMENTAL[0]}",
        f"--{ATBU_BACKUP_TYPE_INCREMENTAL}",
        help="""Perform an incremental backup. Files whose modified date, size in bytes have changed
since the prior backup.
""",
        action="store_true",
    )
    group_backup_operation.add_argument(
        f"--{ATBU_BACKUP_TYPE_INCREMENTAL_PLUS_SHORT_CMD_OPT}",
        f"--{ATBU_BACKUP_TYPE_INCREMENTAL_PLUS}",
        help="""Perform a comprehensive incremental backup. Files whose modified date, size in bytes,
or hash have changed since the prior backup. This reqiures generating the hash of all
source files which can be intensive and time consuming. Generally, it may be a good
idea, depending on the number of source files, to perform an incremental plus
backup once a while to ensure maximum integrity.
""",
        action="store_true",
    )
    parser_backup.add_argument(
        "source_dirs",
        metavar="source-dirs",
        type=str,
        nargs="+",
        help=f"""One more locations to backup.""",
    )
    parser_backup.add_argument(
        "dest_storage_specifier",
        metavar="dest-storage-specifier",
        type=str,
        help=f"""The destination to backup. The destiation is a <storage-def-specifier>.
To learn more about specifying storage definitions, backups, and files,
see '{ATBU_PROGRAM_NAME} help specifiers' for details.
""",
    )
    parser_backup.add_argument(
        "--dedup",
        choices=[
            ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST,
            ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST_EXT,
        ],
        nargs=1,
        help=f"""For --incremental-plus backups only, employ deduplication given the designated
deduplication method:

    '{ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST}':     This option causes the backup to consider a file unchanged if its
                  digest/date/size match that of an already backed up file, regardless
                  of the other file's path or filename.

    '{ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST_EXT}': This option is the same as '{ATBU_BACKUP_DEDUPLICATION_TYPE_DIGEST}' except a file must have an extension
                  which matches the already backed up file in order to be considered a
                  duplicate of that file.
                  
                  In all cases, if a deduplication match is detected, the file being
                  considered for backup not backed up but is instead marked a duplicate.
                  During restore, a duplicate resolution process will take place,
                  re-associating a duplicate with a physically backed up file matching
                  the same characteristics used to determine it was a duplicate. Generally
                  speaking, a {DEFAULT_HASH_ALGORITHM} hash match will very rarely have a collision. The
                  additional checks, such as date/size/extension are additional checks
                  which may help mitigate a rare hash collision.

""",
    )
    parser_backup.add_argument(
        "-e",
        "--exclude",
        default=[],
        nargs="+",
        help=f"""Specify one or more glob-like patterns that, if any are matched to a file's path,
the file will not be included in the backup.
""",
    )
    parser_backup.add_argument(
        "--detect-bitrot",
        "--dbr",
        "--detect-sneaky-corruption",
        "--dsc",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""Enabled by default, reports potential bitrot (aka "sneaky corruption") as an error.
Requires use of --incremental-plus (--ip). Use --no-detect-bitrot (--no-dbr) to
squelch reporting as an error (still reported informationally). Bitrot or so-called
sneaky corruption is detected when a file's date/time and size have remained the
same since the last backup, but the digests are different.""",
    )
    parser_backup.add_argument(
        "-z",
        "--compression",
        choices=BACKUP_COMPRESSION_CHOICES,
        help=f"""Set the backup compression level. The default is '{BACKUP_COMPRESSION_DEFAULT}'.
""",
    )
    parser_backup.set_defaults(func=handle_backup)

    #
    # 'restore' subparser.
    #
    parser_restore = subparsers.add_parser(
        "restore",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parser_common, parser_credential_filename_optional],
        help="Restore selected files from a backup.",
        description=f"""Restore selected files from a backup.

Examples:

    {ATBU_PROGRAM_NAME} restore d:\\MySourceBackup backup:last files:* c:\\MyRestoreDirectory
        From the last backup to d:\\MySourceBackup, restore all files to c:\\MyRestoreDirectory
        In this case, d:\\MySourceBackup is a local backup (located on d: drive) and not a cloud backup.

    {ATBU_PROGRAM_NAME} restore storage:my_cloud_backup backup:last files:* c:\\MyRestoreDirectory
        From the last backup to storage:my_cloud_backup, restore all files to c:\\MyRestoreDirectory
        In this case, storage:my_cloud_backup is a cloud backup.    
""",
    )
    parser_restore.add_argument(
        "source_storage_specifiers",
        metavar="source-storage-specifiers",
        type=str,
        nargs="+",
        help=f"""The source selections to restore from. To learn more about specifying storage definitions,
backups, and files, see '{ATBU_PROGRAM_NAME} help specifiers' for details.
""",
    )
    parser_restore.add_argument(
        "restore_dir",
        metavar="restore-dir",
        type=str,
        help=f"""The destination directory for the restored files.""",
    )
    parser_restore.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""If specified, allows overwriting already existing destination files.""",
    )
    parser_restore.add_argument(
        "--auto-mapping",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="""Use --no-auto-mapping to disable automatic restore path mapping. By default,
paths mapped to allow restoring with minimal path components relative
to the restore directory. When that is not desirable, use --no-auto-mapping.
""",
    )
    parser_restore.set_defaults(func=handle_restore)

    #
    # 'verify' subparser.
    #
    parser_verify = subparsers.add_parser(
        "verify",
        help="Verify selected files from a backup.",
        description=f"""Verify selected files from a backup.

Examples:

    {ATBU_PROGRAM_NAME} verify d:\\MySourceBackup backup:last files:*
        From the last backup to d:\\MySourceBackup, verify all files.
        In this case, d:\\MySourceBackup is a local backup (located on
        d: drive) and not a cloud backup.

    {ATBU_PROGRAM_NAME} verify storage:my_cloud_backup backup:last files:*
        From the last backup to storage:my_cloud_backup, verify all files
        In this case, storage:my_cloud_backup is a cloud backup.

To learn more about specifying a storage definitions, backups,
and files, see '{ATBU_PROGRAM_NAME} help specifiers'
""",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parser_common, parser_credential_filename_optional],
    )
    parser_verify.add_argument(
        "source_storage_specifiers",
        metavar="source-storage-specifiers",
        type=str,
        nargs="+",
        help=f"""The source selections to restore from. To learn more about specifying storage definitions,
backups, and files, see '{ATBU_PROGRAM_NAME} help specifiers' for details.
""",
    )
    parser_verify.add_argument(
        "--compare",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""
If specified, performs comparison with local copy of the file. If not specified, only the digests
are calculated/compared (default).
""",
    )
    parser_verify.add_argument(
        "--compare-root",
        type=str,
        help="""
When --compare is specified, --compare-root, if specified, is folder under which are the
local files used for comparison. If --compare-root is not specified, the file's original
backup location is used.
""",
    )
    parser_verify.set_defaults(func=handle_verify)

    #
    # 'list' subparser.
    #
    parser_list = subparsers.add_parser(
        "list",
        help="List storage definitions, backups, and files relating to backups.",
        parents=[parser_common, parser_credential_filename_optional],
    )
    parser_list.add_argument(
        "specifiers",
        metavar="[<source> <specific-backup> <file-selections>]",
        type=str,
        nargs="*",
        help=f"""
The source backup specifiers for the list.
Examples:
    {ATBU_PROGRAM_NAME} list
        List all known storage definitions.
    {ATBU_PROGRAM_NAME} list d:\\MySourceBackup backup:last files:*
        From the last backup to d:\\MySourceBackup, list all files.
        In this case, d:\\MySourceBackup is a local backup (located on d: drive) and not a cloud backup.
    {ATBU_PROGRAM_NAME} list storage:my_cloud_backup backup:last files:*
        From the last backup to storage:my_cloud_backup, list all files
        In this case, storage:my_cloud_backup is a cloud backup.
To learn more about specifying a storage definitions, backups, and files, see '{ATBU_PROGRAM_NAME} help specifiers'
""",
    )
    parser_list.set_defaults(func=handle_list)

    #
    # Common storage-def-specifier argument for creds subparsers.
    #
    parser_creds_storage_def_specifier = argparse.ArgumentParser(add_help=False)
    parser_creds_storage_def_specifier.add_argument(
        "storage_def",
        metavar="<storage-def-specifier>",
        type=str,
        help=f"""Backup storage definition to affect. This can be a storage definition name or file
path to file system storage location. See '{ATBU_PROGRAM_NAME} help specifiers'.
""",
    )

    #
    # 'creds' subparser.
    #
    parser_creds = subparsers.add_parser(
        "creds",
        formatter_class=argparse.RawTextHelpFormatter,
        help=f"""Create/delete backup configurations, backup/restore/set credentials.""",
        parents=[parser_common],
    )
    subparser_creds = parser_creds.add_subparsers(
        dest="subcmd",
    )

    #
    # creds create-storage-def:
    #

    # pylint: disable=unused-variable
    cred_create_storage_def_parser = subparser_creds.add_parser(
        CREDS_SUBCMD_CREATE_STORAGE_DEF,
        parents=[parser_creds_storage_def_specifier, parser_credential_filename_optional, parser_common],
        formatter_class=argparse.RawTextHelpFormatter,
        help=f"Create a backup storage definition.",
        description=f"""where

    <interface>           <'filesystem','libcloud'|'google'>
    <provider>            <'filesystem'|'azure_blobs'|'google_storage'>
    <container>           The cloud storage container or bucket name.
    <key>                 access key
    <secret_access_key>   storage secret access key
    [,k=v[,k=v ...]]      extra parameters as needed: name1=value1,name2=value2,...

Create a storage definition. No spaces allowed as shown. If <secret_access_key> evaluates to an
existing filename with a .json extension, the path to that file will be saved as the crednetial
(i.e., an OAuth2 service account file). For some configurations, it may be easier or necessary
to edit the config .json file directly, adding whatever driver arguments you wish.

Examples:

    {ATBU_PROGRAM_NAME} creds {CREDS_SUBCMD_CREATE_STORAGE_DEF} my-cloud-backup libcloud azure_blobs key=<access_key>,secret=<secret_access_key>
    {ATBU_PROGRAM_NAME} creds {CREDS_SUBCMD_CREATE_STORAGE_DEF} my-cloud-backup google google_storage key=<client_email>,secret=<path_to_OAuth2.json>

""",
    )
    cred_create_storage_def_parser.add_argument(
        "interface",
        choices=[
            CONFIG_INTERFACE_TYPE_FILESYSTEM,
            CONFIG_INTERFACE_TYPE_LIBCLOUD,
            CONFIG_INTERFACE_TYPE_GOOGLE,
            CONFIG_INTERFACE_TYPE_AZURE,
        ],
        help="""The interface (API) the storage definition should use.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "provider",
        metavar="<libcloud_provider_id>",
        help=f"""The storage provider name. This can be '{CONFIG_INTERFACE_TYPE_FILESYSTEM}' for a local
directory, or a provider supported by the chosen interface. For a cloud provider, if
interface is '{CONFIG_INTERFACE_TYPE_LIBCLOUD}' you will typically specify a libcloud provider string such as
'azure_blobs' or 'google_storage'. For interface '{CONFIG_INTERFACE_TYPE_GOOGLE}' use provider 'google_storage'.
Interface '{CONFIG_INTERFACE_TYPE_AZURE}' allows us of either traditional S3 secret or SAS token as secret. For
interface '{CONFIG_INTERFACE_TYPE_AZURE}' use provider 'azure_blobs'.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "container",
        metavar="<container-name>",
        help=f"""The container name to use. If the name ends with '*', then this parameter acts as a prefix
in the search for an available name which, if found, is then created. A unique
UUID is appended to achieve the goal of finding a name. If a single name without '*'
is specified, it must either already exist or you can use --create-container to
attempt to create that one name.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "driver_params",
        metavar="key=<access_key>,secret=<secret_access_key>[,k=v[,k=v ...]]",
        nargs="?",
        default=None,
        help=f"""The parameters for the API driver. For example, these are passed to libcloud to
create a driver to access the API. Generally, you must specify at least a key and secret, but
you can specify any libcloud driver arguments.

""",
    )
    cred_create_storage_def_parser.add_argument(
        "--create-container",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""Attempt to create the container. If you do not specify this, you must
use your cloud provider's user interface to create the container.
""",
    )
    cred_create_storage_def_parser.add_argument(
        "--include-iv",
        "--iiv",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""The default is --include-iv which stores the encryption initialization vector (IV) in the
clear with the encrypted file in the backup storage location. It is recommended you do not
override the default without which you cannot recover encrypted files should you
fail to properly backup your locally stored backup information (which includes the IVs).
""",
    )
    cred_create_storage_def_parser.add_argument(
        "--secrets-visible",
        "--sv",
        action=argparse.BooleanOptionalAction,
        default=False,
        help=f"""When <driver_params> is not specified on the command line, the user is prompted for input.
This switch causes user input of secrets (i.e., access/secret keys) to be visible while typing.
"""
    )

    #
    # creds delete-storage-def:
    #
    cred_delete_storage_def_parser = subparser_creds.add_parser(
        "delete-storage-def",
        help="Delete a storage definition.",
        parents=[parser_creds_storage_def_specifier, parser_common],
    )
    cred_delete_storage_def_parser.add_argument(
        "--force",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="Delete without a confirmation prompt.",
    )
    cred_delete_storage_def_parser.add_argument(
        "--delete-backup-info",
        "--dbi",
        action=argparse.BooleanOptionalAction,
        help="""If you specify --no-delete-backup-info (--no-dbi), backup information files
will not be deleted. By default, a storage definition's backup informatiion
files are deleted along with the storage definition itself. If this option
is used without --force, prompting for deleting backup information files
will be skipped (i.e., user only prompted once re storage defintiion).
""",
    )

    #
    # creds export:
    #
    cred_export_cred_parser = subparser_creds.add_parser(
        "export",
        help="Export a storage definition and its secrets to a text file (WARNING: clear text file!).",
        parents=[
            parser_creds_storage_def_specifier,
            parser_credential_filename_positional,
            parser_file_ovewrite,
            parser_common,
        ],
    )

    #
    # creds import:
    #
    cred_import_cred_parser = subparser_creds.add_parser(
        "import",
        help="Import a storage definition and its secrets from a text file.",
        parents=[
            parser_creds_storage_def_specifier,
            parser_credential_filename_positional,
            parser_common,
        ],
    )
    # TODO: Remove if ultimately unused:
    cred_import_cred_parser.add_argument(
        "--create-config",
        action=argparse.BooleanOptionalAction,
        default=False,
        help=f"""Create a {ATBU_USER_DEFAULT_CONFIG_FILE_NAME} configuration if one does not exist.""",
    )
    cred_import_cred_parser.add_argument(
        "--prompt",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""
If --no-prompt is specified, prompts will be skipped. If you specify
to import a backed up configuration, if the same configuration is
already present, it will be overwritten. Ovewriting will wipe out any
existing keys for that storage definition. If keys are different from
the imported configuration, you could lose overwritten keys. It is
recommended you backup configurations before overwriting them by using
'{ATBU_PROGRAM_NAME} creds export ...' first.
""",
    )

    #
    # creds set-password:
    #
    creds_set_password_parser = subparser_creds.add_parser(
        name=CRED_OPERATION_SET_PASSWORD,
        aliases=[CRED_OPERATION_SET_PASSWORD_ALIAS],
        formatter_class=argparse.RawTextHelpFormatter,
        help="Set password for the backup or storage.",
        parents=[parser_creds_storage_def_specifier, parser_common],
    )
    creds_set_password_parser.add_argument(
        "password_type",
        choices=CRED_OPERATION_SET_PASSWORD_TYPES,
        help=f"""What password type or secret to set (i.e., backup password, cloud storage secret, etc.).
    '{CRED_OPERATION_SET_PASSWORD_TYPE_BACKUP}' : Set the password for the backup as a whole. This password is used
        to encrypt both the backup encryption key and/or storage secret.
    '{CRED_OPERATION_SET_PASSWORD_TYPE_STORAGE}' : Set the storage secret for the cloud backup storage definition.
    '{CRED_OPERATION_SET_PASSWORD_TYPE_FILENAME}' : Set the storage secret to an OAuth2 .json file path name.*
    '{CRED_OPERATION_SET_PASSWORD_TYPE_ENVVAR}' : Set the storage secret to an environment variable name that itself
        points to an OAuth2 .json filename.*
    * OAuth2 .json files in format used with a GCS service account are supported.
""",
    )
    creds_set_password_parser.add_argument(
        "password",
        help=f"""The password value to set. If not specified, you will be prompted. The following
are the potential values based on your selection for the above password type argument:
    '{CRED_OPERATION_SET_PASSWORD_TYPE_BACKUP}' : Specify a textual string password used to encrypt the backup encryption key.
    '{CRED_OPERATION_SET_PASSWORD_TYPE_STORAGE}' : Specify a storage secret (i.e., what you copy/paste from cloud portal).
    '{CRED_OPERATION_SET_PASSWORD_TYPE_FILENAME}' : Specify a path to a GCS OAuth2 service account .json file.
    '{CRED_OPERATION_SET_PASSWORD_TYPE_ENVVAR}' : Specify an environment variable that itself points to a GCS OAuth2
        service account .json file.
""",
        nargs="?",
    )
    parser_creds.set_defaults(func=handle_creds)

    #
    # 'recover' subparser.
    #
    parser_recover = subparsers.add_parser(
        "recover",
        parents=[parser_common, parser_credential_filename_optional],
        formatter_class=argparse.RawTextHelpFormatter,
        help="""Recover your backup information files.""",
        description=f"""
This option allows you to recovery backup information from the backup storage, cloud or
local file system. You would do this if your local information has become lost, corrupted
or is otherwise unavailable. You still require your backup's private key, so you should
always make sure you backup (export) your backup's configuration and store it in a safe
place. Without that, you cannot use this recover option.

For more info on exporting your configuration, see '{ATBU_PROGRAM_NAME} creds export ...'
""",
    )
    parser_recover.add_argument(
        "storage_def_cred_cfg",
        metavar="<exported-config-filename> | <storage-def-specifier> [exported-config-filename]",
        nargs="+",
        type=str,
        help=f"""
You can specify the exported config file alone if the recovery
is for a cloud provider storage as the file itself has the
storage definition name, keys etc. needed for recovery.

If you just wish to recover backup information for an already
existing storage def configuration, specify the storage def
name alone, or a path to a filesystem storage definition path.

If you wish to both import credentials and recover backup
information for a filesystem storage, specify the definition's
path and the config filename to import.

The storage definition/credential backup config file is
created by exporting using '{ATBU_PROGRAM_NAME} creds export ...'. For good
measure, you should always backup <home>/{ATBU_DEFAULT_CONFIG_DIR_NAME}/{ATBU_DEFAULT_BACKUP_INFO_SUBDIR}
before performing backup info recovery in case you are uncertain
what you are doing, or otherwise later desire to use or look at
what you had before recovery.
""",
    )
    parser_recover.add_argument(
        "--prompt",
        action=argparse.BooleanOptionalAction,
        default=True,
        help=f"""
If --no-prompt is specified, prompts will be skipped. If you specify
to import a backed up configuration, if the same configuration is
already present, it will be overwritten. Ovewriting will wipe out any
existing keys for that storage definition. If keys are different from
the imported configuration, you could lose overwritten keys. It is
recommended you backup configurations before overwriting them by using
'{ATBU_PROGRAM_NAME} creds export ...' first.
""",
    )
    parser_recover.set_defaults(func=handle_recover)

    #
    # 'decrypt' subparser.
    #
    parser_decrypt = subparsers.add_parser(
        "decrypt",
        formatter_class=argparse.RawTextHelpFormatter,
        parents=[parser_common, parser_credential_filename_optional],
        help="Decrypt storage files directly.",
        description=f"""Decrypt storage files directly. This is generally a command used by someone with technical expertise who may have
a need to decrypt backup storage files without needing to download them. For normal usage/users, this command can
be ignored.

This command might be used in a case where you have the direct raw storage files and you wish to decrypt them. For
example, if you were to download or otherwise get a copy of all your encrypted files stored in the cloud, where
downloading was not required, but you needed to decrypt those files locally in your environment.

Another example might be to sanity check that you can decrypt your storage files without requiring download, perhaps
for prep and confidence that the first example above can be achieved. This might be performed as part of disaster
recovery preparation.

To use this command, you still require your encryption private key so a storage definiton with that key needs to be
configured.

Examples:

    {ATBU_PROGRAM_NAME} decrypt storage:my_cloud_backup c:\\EncryptedFiles c:\\RestoreLocation
        This would use the private encryption key of my_cloud_backup to decrypt files located
        in c:\\EncryptedFiles to c:\\RestoreLocation.
""",
    )
    parser_decrypt.add_argument(
        "private_key_storage_specifier",
        metavar="private-key-storage-specifier",
        type=str,
        help=f"""The storage definition whose private key should be used for decryption.
""",
    )
    parser_decrypt.add_argument(
        "source_files_dir",
        metavar="source-files-dir",
        type=str,
        help=f"""The source directory with optional glob-like pattern specifying the files to decrypt.
If no glob pattern is present, ** is used to select all files. Regardless of the files
found, only those files ending with either {ATBU_FILE_BACKUP_EXTENSION_ENCRYPTED} or {ATBU_FILE_BACKUP_EXTENSION} are considered.""",
    )
    parser_decrypt.add_argument(
        "restore_dir",
        metavar="restore-dir",
        type=str,
        help=f"""The destination directory for the restored files.""",
    )
    parser_decrypt.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""If specified, allows overwriting already existing destination files.""",
    )
    parser_decrypt.set_defaults(func=handle_decrypt)

    #### decrypt END

    #############################################################################################
    #                           Persistence-related argparse setup                              #
    #############################################################################################

    #
    # Common to subparsers saving/overwriting .atbu/diff databases.
    #

    #
    # Common to update and diff subparsers.
    #
    common_change_detection_type = argparse.ArgumentParser(add_help=False)
    common_change_detection_type.add_argument(
        "--change-detection-type",
        "--cdt",
        choices=CHANGE_DETECTION_CHOICES,
        default=CHANGE_DETECTION_TYPE_DATESIZE,
        help=f"""For operations where digests can be updated, this option specifies the
method to use when determining if a file has changed, where persistent
info needs updating. The options are:
    '{CHANGE_DETECTION_TYPE_DATESIZE}' for file system date or size changes (default).
    '{CHANGE_DETECTION_TYPE_DIGEST}' for digest check (requires time/cpu for re-gen of file digest).
    '{CHANGE_DETECTION_TYPE_FORCE}' will for an update to all files without the need for checks.
""",
    )
    common_update_stale = argparse.ArgumentParser(add_help=False)
    common_update_stale.add_argument(
        "--update-stale",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="""Specifying --no-update-stale disables the default of updating stale
persistent information. If you specify --no-update-stale, the operation
(such as diff) may be performed using stale persisted information which
may or may not be acceptable depending on your situation. Generally, the
default is what you want because it updates stale information when a
change is detected as determined by --change-detection-type.""",
    )

    #
    # Update digests subparser.
    #
    parser_update_digests = subparsers.add_parser(
        "update-digests",
        parents=[common_change_detection_type, parser_common],
        formatter_class=argparse.RawTextHelpFormatter,
        help="Update persistent file information in the specified directories.",
        description=f"""
Generate or update digests for all files in specified directories, placing the persistent
information in sidecar {ATBU_PERSISTENT_INFO_EXTENSION} files.

WARNING: when using the --persist-type option 'per-file' or 'both', this command creates a new
<filename>{ATBU_PERSISTENT_INFO_EXTENSION} file for all encountered files. By contrast, the 'per-dir' option (the default)
creates a single database at the root of the specified location.

""",
    )
    parser_update_digests.add_argument(
        "locations",
        metavar="[per-file:|pf:|per-dir:|pd:|per-both:|pb:]|<location>",
        nargs="+",
        type=str,
        help=f"""One or more local file system directories whose persistent information to update,
each optionally prefixed with the desired file info persistence to use (default is per-dir).
Note, there is a space after per-file:, per-dir:, and any location. For example,
'per-file: c:\\SomeLocationThatShouldUsePerFile'.
""",
    )
    parser_update_digests.set_defaults(func=handle_update_digests)

    #
    # 'diff' subparser.
    #
    parser_diff = subparsers.add_parser(
        "diff",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Perform a diff of persistent file information of two different directories.",
        description="""Perform a compare of two locations, location A and B, each which has persistent information. Either
location A or B can be a directory or a persistent file information database. The diff produces a
result of A less B, meaning files A that are not in B. Optionally, you can choose to have an action
of removing or moving discovered duplicates.
""",
        parents=[parser_common, common_update_stale, common_change_detection_type],
    )
    parser_diff.add_argument(
        "locations",
        metavar="[per-file:|pf:|per-dir:|pd:|per-both:|pb:]|<location>",
        nargs="+",
        type=str,
        help=f"""You must specify this twice, once each for location A and B, optionally prefixing
either with the desired file info persistence to use (default is per-dir).
Note, there is a space after per-file:, per-dir:, and any location. For example,
'per-file: c:\\SomeLocationThatIsUsingPerFile'.
""",
    )
    parser_diff.add_argument(
        "--relpath-match",
        action="store_true",
        default=False,
        help=f"""If specified, files are only considered to match between location A and
location B if their relative paths match. Relative paths are derived by
removing the top-level location path under which a file was found. By
default, relative paths do not need ot match, only the digest between
the two locations must patch.
""",
    )
    parser_diff.add_argument(
        "--action",
        choices=DIFF_COMMAND_CHOICES,
        help=f"""Perform an action, either remove or move, on files in location A for
which there are duplicates in location B.""",
    )
    parser_diff.add_argument(
        "--move-destination",
        "--md",
        type=str,
        help="Directory for destination when moving duplicates as part of diff.",
    )
    parser_diff.add_argument(
        "--whatif",
        action="store_true",
        default=False,
        help="Show what files would be removed but do not actually remove them.",
    )
    parser_diff.set_defaults(func=handle_diff)

    #
    # 'savedb' subparser.
    #
    parser_savedb = subparsers.add_parser(
        "save-db",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Save persistent file information from or more directories into a single json db.",
        description=f"""Save a json database from all persistent information created by the '{ATBU_PROGRAM_NAME} update-digests' command.

Console/logging output includes information about any files detected to have changed since the last
update. Use -v to include console/logging of each path checked. Use -vv to include details for each
path even if not changed since the last update. Details are always output when updates occur, where
using -v or -vv will slow down the command considerably for large directories.

""",
        parents=[parser_common, common_update_stale, common_change_detection_type],
    )
    parser_savedb.add_argument(
        "locations",
        metavar="[per-file:|pf:|per-dir:|pd:|per-both:|pb:]|<location>",
        nargs="+",
        help=f"""The locations whose persistent file information should be placed into a newly
saved database .json file. Optionally prefix any location with the desired
file info persistence to use (default is per-dir). Note, there is a space after
per-file:, per-dir:, and any location. For example,
    'per-file: c:\\SomeLocationThatShouldUsePerFile'.
""",
    )
    parser_savedb.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="When requesting to write a hash db file, overwrite any existing db file.",
    )
    parser_savedb.add_argument(
        "--database",
        "--db",
        type=str,
        required=True,
        help="Path of the persistent file information json database.",
    )
    parser_savedb.set_defaults(func=handle_savedb)

    #
    # 'arrange' subparser.
    #
    parser_arrange = subparsers.add_parser(
        "arrange",
        formatter_class=argparse.RawTextHelpFormatter,
        help="Arrange \"target\" files to match a \"template\" directory's structure.",
        description="""Given two directories that were once mirroed, arrange the target
so that any existing target files matching those in the template directory are moved to a new
target directory structure to maximize the match between the two. This can be used to arrange
the target's stale structure to match a newly updated structure in the template directory.
""",
        parents=[parser_common, common_update_stale, common_change_detection_type],
    )
    parser_arrange.add_argument(
        "locations",
        metavar="[per-file:|pf:|per-dir:|pd:|per-both:|pb:] <template_root> [pf:|pd:|pb:] <target_source_root> [pf:|pd:|pb:] <target_dest_root>",
        nargs="+",
        type=str,
        help=f"""You must specify 3 root path locations: template root, target source, and target destination.
The target source and destination must be on the same drive, where the arrange command is for
rearranging (moving) files on that drive (from source to dest) based on the template root structure.

You can specify a database type for each location by using the appropriate per-file, per-dir,
or per-both setting prior to the location(s) it applies to. For large media files, or even
otherwise, it is recommended to use per-file even though it's no the default because it
creates sidecar files. The per-dir approach is currently for one-off operations or where
sidecar files are undesired. Generally, for large media management with sidecar .atbu files,
specify 'pf:' before the first location.

The arrange command will use the template root path to arrange (move) files within the target
source to the target destination, where the placement within the target destination will match
the template root's arrangement for any files available and matching within the target source.

Example: C:\\root\\a\\file1.txt is manually mirrored to D:\\root\\a\\file1.txt, where file1.txt on C:
is moved to C:\\root\\b\\file1.txt while D: is offline. Their structures are now out of sync.
When D: is brought back online (reconnected via USB or whatever), its desired to arrange D: to
match structural changes to C: as best as possible, including renames which move and do not move
the folder location of a file. Using arrange with template C:\\root, target source D:\\root, and
target destination D:\\rootnew, results in a best-effort to move files from D:\\root to D:\\rootnew
while using C:\\root as a template for the desired names/structure. In this example case, arrange
results in D:\\root\\a\\file1.txt being moved to D:\\rootnew\\b\\file1.txt. After such a rearrangement,
normal copy-style manual mirroring can resume.
""",
    )
    parser_arrange.add_argument(
        "--no-undo",
        action="store_true",
        default=False,
        help="""Must be specified if --undofile is not specified. This is a sanity check to ensure",
no undo file is truely desired.
"""
    )
    parser_arrange.add_argument(
        "-u", "--undofile",
        type=str,
        default=None,
        help="""Path to write the arrange undo file which contains information about the operations
performed which can be used to undo the arranging. Note, at this time the undo file is not used
by this tool but eventually will be. For the time being, it can be used by scripting languages
to undo move operations if desired, or to otherwise have a log of the operations which took
place.
""",
    )
    parser_arrange.add_argument(
        "--overwrite",
        action=argparse.BooleanOptionalAction,
        default=False,
        help="""If specified, existing target destination files can be ovewritten. By default,
move operations will fail if the destination file already exists.""",
    )
    parser_arrange.add_argument(
        "--whatif",
        action="store_true",
        default=False,
        help="Show how files would be arranged without actually arranging them.",
    )
    parser_arrange.set_defaults(func=handle_arrange)

    #############################################################################################
    #                specific help argparse setup (i.e., for atbu help <subject>                #
    #############################################################################################

    parser_help = subparsers.add_parser(
        "help",
        formatter_class=argparse.RawTextHelpFormatter,
        help=f"""Show help on specific subjects:
    '{ATBU_PROGRAM_NAME} help -h' to see available subjects.""",
    )
    parser_help.add_argument(
        "help_subject",
        choices=[k for k in extra_help_subjects],
        nargs=1,
        help=f"""The subject to get detailed help on:
    '{ATBU_PROGRAM_NAME} help specifiers' - get detailed help on storage definition specifiers used by backup/restore commands.
""",
    )

    return parser


def wait_for_debugger_attach(debug_server_port: Union[str, int, list]):
    if debug_server_port is None:
        raise InvalidCommandLineArgument(f"The debugger port must be specified.")
    if isinstance(debug_server_port, list):
        debug_server_port = debug_server_port[0]
    if isinstance(debug_server_port, str):
        try:
            debug_server_port = int(debug_server_port)
        except Exception as ex:
            raise InvalidCommandLineArgument(
                f"Parsing error: "
                f"The value '{debug_server_port}' is an invalid port number."
            ) from ex
    if debug_server_port < 0 or debug_server_port > 65535:
        raise InvalidCommandLineArgument(
            f" The value '{debug_server_port}' is an invalid port number."
        )
    try:
        import debugpy  # pylint: disable=import-outside-toplevel

        debugpy.listen(debug_server_port)
        print(f"Waiting for the debugger to attach via port {debug_server_port}...")
        debugpy.wait_for_client()
        debugpy.breakpoint()
        print(f"Debugger connected.")
    except ModuleNotFoundError as ex:
        raise InvalidStateError(
            f"Cannot 'import debugpy'. Either ensure vscode debugpy is available."
        ) from ex
    except Exception as ex:
        raise InvalidStateError(
            f"Unexpected error. Cannot wait for debugger attach. {exc_to_string(ex)}"
        ) from ex


def process_profile_arguments(args) -> EasyProfile:
    easy_profiler = None
    if args.profile:
        profile_file = None
        profile_stats = args.profile_stats
        if hasattr(args, "profile_file") and args.profile_file is not None:
            if os.path.exists(args.profile_file):
                raise InvalidCommandLineArgument(
                    f"The --profile-file file name must not already exist: "
                    f"file={args.profile_file}"
                )
            profile_file = args.profile_file
        if not profile_stats and profile_file is None:
            raise InvalidCommandLineArgument(
                f"Either one or both of --profile-file / --profile-stats must be specified "
                f"when --profile is specified."
            )
        easy_profiler = EasyProfile(log_stats=profile_stats, profile_file=profile_file)
        easy_profiler.start()
    return easy_profiler


def main(argv=None):
    multiprocessing.set_start_method("spawn")
    global_init()
    initialize_logging_basic()
    # pdb.set_trace()
    logging.info(f"{ATBU_ACRONYM} - v{ATBU_VERSION_STRING}")

    parser = create_argparse()
    args = parser.parse_args(argv)

    if hasattr(args, "debug_server") and args.debug_server is not None:
        wait_for_debugger_attach(args.debug_server)

    if hasattr(args, "help_subject"):
        for requested_help_subject in args.help_subject:
            if requested_help_subject in extra_help_subjects:
                print(extra_help_subjects[requested_help_subject])
                parser.exit(1)
        raise InvalidStateError(
            f"Expected help subject to match those available: {args.help_subject}"
        )

    if hasattr(args, "automated_testing"):
        set_automated_testing_mode(args.automated_testing)

    verbosity_level = 0
    if hasattr(args, "verbosity") and args.verbosity is not None:
        verbosity_level = args.verbosity

    # Instantiate global singleton GlobalHasherDefinitions.
    try:
        GlobalHasherDefinitions([DEFAULT_HASH_ALGORITHM])
    except SingletonAlreadyCreated:
        logging.warning(f"GlobalHasherDefinitions already created.")

    logfile = None
    loglevel = None
    if hasattr(args, "logfile"):
        logfile = args.logfile
    if hasattr(args, "loglevel"):
        loglevel = args.loglevel
    # If no log file specified and user is creating/updating database,
    # automatically place log file side-by-side with db.
    if (
        not logfile
        and hasattr(args, "func")
        and args.func == handle_savedb  # pylint: disable=comparison-with-callable.
        and hasattr(args, "database")  # pylint: disable=W0143
        and args.database is not None
    ):
        logfile = args.database + ".log"

    log_console_detail = False
    if hasattr(args, "log_console_detail") and args.log_console_detail:
        log_console_detail = args.log_console_detail

    exit_code = 1
    try:
        if hasattr(args, 'config_file') and args.config_file is not None:
            if not os.path.isfile(args.config_file):
                raise CredentialSecretFileNotFoundError(
                    f"The credential configuration file was not found: {args.config_file}",
                )
            #
            # Register the storage definition configuration if one is not already present.
            #
            register_storage_def_config_override(
                storage_def_config_path=args.config_file,
                only_if_not_already_present=True,
            )

        if hasattr(args, "func"):
            remove_created_logging_handlers()
            remove_root_stream_handlers()
            initialize_logging(
                logfile=logfile,
                loglevel=loglevel,
                verbosity_level=verbosity_level,
                log_console_detail=log_console_detail,
            )
            debug_argv = argv if argv is not None else sys.argv
            logging.debug(f"argv={'None' if debug_argv is None else ' '.join([*debug_argv])}")

            if get_automated_testing_mode():
                logging.info(f"*** Automated testing mode is enabled. ***")

            if hasattr(args, "yk"):
                if not isinstance(args.yk, bool):
                    raise InvalidStateError(f"--yk should yield a bool option.")
                if args.yk:
                    set_require_yubikey(is_required=args.yk)
                    try:
                        setup_yubikey_infra()
                    except YubiKeyBackendNotAvailableError as ex:
                        logging.error(
                            f"YubiKey Manager is not installed or "
                            f"could not be initialized."
                        )
                        raise

            profile = process_profile_arguments(args)

            exit_code = args.func(args)
            if exit_code is None:
                exit_code = 0

            if profile is not None:
                profile.stop()
                profile = None
        else:
            print(f"I have nothing to do. Try {ATBU_PROGRAM_NAME} -h for help.")
    except AtbuException as err:
        logging.error(f"Failed: {err.message}")
        if get_verbosity_level() > 0:
            logging.error(exc_to_string(err))
            raise
    finally:
        try:
            deinitialize_logging()
        except QueueListenerNotStarted:
            pass
        except Exception as ex:
            print(f"Failure during deinitialize_logging. {exc_to_string(ex)}")
    logging.debug(f"{ATBU_PROGRAM_NAME} exit_code={exit_code}")
    return exit_code


if __name__ == "__main__":
    main()



# Copyright 2022 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""ATBU interface for Azure Blob Storage via azure-storage-blob portion
of the Azure SDK. Using azure-storage-blob allows for use of the more
robust Shared Access Signature (SAS) credentials.
"""

# pylint: disable=missing-class-docstring)

from typing import Iterator
from uuid import uuid4

from azure.storage.blob import (
    BlobServiceClient,
    ContainerClient,
    BlobClient,
)
from azure.core.exceptions import (
    ResourceExistsError,
    ServiceRequestError,
)

from atbu.common.exception import (
    exc_to_string,
    ObjectDoesNotExistError,
    InvalidContainerNameError,
    ContainerAlreadyExistsError,
    ContainerAutoCreationFailed,
)

from ..credentials import CredentialByteArray
from .base import (
    AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR,
    StorageContainerInterface,
    StorageInterface,
    StorageObjectInterface,
)
from ..constants import *
from ..config import *

DEFAULT_AZURE_RETRY_EXCEPTIONS = (
    ServiceRequestError,
)

class AzureStorageObjectInterface(StorageObjectInterface):
    def __init__(self, storage_object: BlobClient):
        super().__init__()
        self.storage_object = storage_object

    @property
    def size(self):
        return self.storage_object.get_blob_properties().size

    @property
    def hash(self):
        return self.storage_object.get_blob_properties().content_settings.content_md5.hex()

    @property
    def name(self):
        return self.storage_object.blob_name


class AzureStorageContainerInterface(StorageContainerInterface):
    def __init__(self, container: ContainerClient):
        super().__init__()
        self.container = container

    @property
    def name(self) -> str:
        return self.container.container_name

    def get_object(self, object_name: str):
        blob = self.container.get_blob_client(blob=object_name)
        if not blob.exists():
            raise ObjectDoesNotExistError(
                f"The object {object_name} does not exist or was not found."
            )
        object_interface = AzureStorageObjectInterface(storage_object=blob)
        return object_interface

    def delete_object(self, object_name: str):
        blob = self.container.get_blob_client(blob=object_name)
        if blob.exists():
            blob.delete_blob()

    def list_objects(self, prefix: str = None) -> list[StorageObjectInterface]:
        blob_name_list = self.container.list_blob_names(name_starts_with=prefix)
        result: list[StorageObjectInterface] = []
        for bn in blob_name_list:
            blob_client = self.container.get_blob_client(blob=bn)
            result.append(AzureStorageObjectInterface(storage_object=blob_client))
        return result


class AzureStorageInterface(StorageInterface):

    def __init__(self, storage_def):
        super().__init__()
        self._retry_exceptions = DEFAULT_AZURE_RETRY_EXCEPTIONS
        self.storage_def: dict = storage_def
        self.driver_config: dict = self.storage_def[CONFIG_SECTION_DRIVER]
        self.blob_service_client = self._create_blob_service_client()

    @property
    def upload_chunk_size(self):
        max_block_size = 4*1024*1024
        try:
            max_block_size = self.blob_service_client._config.max_block_size
        except:
            pass
        return max_block_size

    @property
    def download_chunk_size(self):
        max_chunk_get_size = 4*1024*1024
        try:
            max_chunk_get_size = self.blob_service_client._config.max_chunk_get_size
        except:
            pass
        return max_chunk_get_size

    def _create_blob_service_client(self):
        password_type = self.driver_config.get(CONFIG_PASSWORD_TYPE)
        if password_type is None:
            raise ValueError(f"Credential type not found in configuration.")
        if password_type != CONFIG_PASSWORD_KIND_ACTUAL:
            raise ValueError(
                f"Expected either an Azure Blob Storage Access Key or SAS."
            )
        # For Azure Blob Storage:
        #   key is the storage account name
        #   secret is either the key (secret) or the SAS token.
        storage_account_name = self.driver_config[CONFIG_VALUE_NAME_DRIVER_STORAGE_KEY]
        # For Azure Blob Storage, the secret is either the access
        secret = self.driver_config[CONFIG_VALUE_NAME_DRIVER_STORAGE_SECRET]
        if isinstance(secret, (CredentialByteArray, bytearray, bytes)):
            secret = secret.decode("utf-8")
        blob_service_client = BlobServiceClient(
            account_url=f"https://{storage_account_name}.blob.core.windows.net",
            credential=secret,
        )
        return blob_service_client

    def get_container(self, container_name: str) -> StorageContainerInterface:
        container_client = self.blob_service_client.get_container_client(container=container_name)
        return AzureStorageContainerInterface(container=container_client)

    def _create_container(self, container_name: str) -> StorageContainerInterface:
        try:
            container_client = self.blob_service_client.create_container(name=container_name)
            return AzureStorageContainerInterface(container=container_client)
        except ResourceExistsError as ex:
            raise InvalidContainerNameError(
                f"Cannot create container '{container_name}' due to a conflict: {exc_to_string(ex)}"
            ) from ex

    def create_container(self, container_name: str) -> StorageContainerInterface:
        if container_name[-1] != AUTO_FIND_CREATE_CONTAINER_INDICATOR_CHAR:
            return self._create_container(container_name=container_name)
        else:
            base_name = container_name[:-1]
            max_attempts = 100
            attempts_remaining = max_attempts
            while True:
                attempts_remaining -= 1
                candidate = f"{base_name}-{str(uuid4())}"
                try:
                    ci = self._create_container(container_name=candidate)
                    return ci
                except (ContainerAlreadyExistsError, InvalidContainerNameError) as ex:
                    if attempts_remaining <= 0:
                        raise ContainerAutoCreationFailed(
                            f"Automatic container creation failed after {max_attempts} attempts. "
                            f"{exc_to_string(ex)}"
                        ) from ex

    def delete_container(self, container_name: str):
        try:
            self.blob_service_client.delete_container(container=container_name)
        except Exception:  # pylint: disable=try-except-raise
            raise  # TBD any customizations.

    def upload_stream_to_object(
        self,
        container: StorageContainerInterface,
        object_name: str,
        stream,
        source_path: str,
    ):
        azure_container: AzureStorageContainerInterface = container
        azure_container.container.upload_blob(
            name=object_name,
            data=stream.get_iterator_only_proxy(),
        )

    def download_object_as_stream(
        self, storage_object: AzureStorageObjectInterface, chunk_size: int
    ) -> Iterator[bytes]:
        ssd = storage_object.storage_object.download_blob()
        for chunk in ssd.chunks():
            c = chunk
            yield chunk



# Copyright 2022, 2023 Ashley R. Thomas
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Persistent file information arranging a target using a template.
"""

import json
import logging
import os
from dataclasses import field


from ..backup.exception import *
from ..backup.constants import ATBU_PERSISTENT_INFO_EXTENSION
from ..backup.global_hasher import GlobalHasherDefinitions
from .file_info import (
    FileInformationPersistent,
    is_file_info_list_bad_state,
)
from .database import (
    FileInformationDatabaseCollection,
    extract_location_info,
    flatten_location_info_to_path_sorted_list,
    rel_path,
)


@dataclass
class _ArrangeOperationInfo:
    template_fi: FileInformationPersistent
    target_source_fi: FileInformationPersistent
    target_source_full_path: str
    target_dest_rel_path: str
    target_dest_full_path: str


@dataclass
class _FileInfoRelPathInfo:

    fi: FileInformationPersistent
    rel_path: str
    rel_path_nc: str = field(init=False)

    def __post_init__(self):
        self.rel_path_nc = os.path.normcase(self.rel_path)


@dataclass
class _ArrangeUndoInfo:

    source_full_path: str
    dest_full_path: str
    is_move_successful: bool
    error_msg: str

    def to_serialization_dict(self) -> dict:
        d = {
            "_type": "_ArrangeUndoInfo",
            "source_full_path": self.source_full_path,
            "dest_full_path": self.dest_full_path,
            "is_move_successful": self.is_move_successful,
            "error_msg": self.error_msg,
        }
        return d

    @staticmethod
    def get_json_encoder() -> json.JSONEncoder:
        class ArrangeUndoInfoEncoder(json.JSONEncoder):
            def default(self, o):
                if isinstance(o, _ArrangeUndoInfo):
                    return o.to_serialization_dict()
                return json.JSONEncoder.default(self, o)
        return ArrangeUndoInfoEncoder


def _build_arrange_operation_list(
    template_root: str,
    template_info: dict[str, list[FileInformationPersistent]],
    target_source_root: str,
    target_source_info: dict[str, list[FileInformationPersistent]],
    target_dest_root: str,
) -> tuple[int, int, int, list[_ArrangeOperationInfo]]:
    primary_hasher_name = GlobalHasherDefinitions().get_primary_hashing_algo_name()

    target_source_fi_list = flatten_location_info_to_path_sorted_list(
        location_info=target_source_info,
    )
    arrange_oper_info_list: list[_ArrangeOperationInfo] = []
    template_match_count = 0
    template_no_match_count = 0
    template_match_exhausted_count = 0
    
    for targ_src_fi in target_source_fi_list:

        # Find template match.
        template_fi_list = template_info.get(targ_src_fi.primary_digest)
        if template_fi_list is None:
            template_no_match_count += 1
            logging.info(
                f"Template match not found for target source: "
                f"{primary_hasher_name}: "
                f"digest={targ_src_fi.primary_digest} "
                f"file_info[0]={targ_src_fi.path}"
            )
            continue

        if len(template_fi_list) == 0:
            template_match_exhausted_count += 1
            logging.warning(
                f"WARNING: Template match found but has been exhausted by other arrange operations: "
                f"{primary_hasher_name}: "
                f"digest={targ_src_fi.primary_digest} "
                f"file_info[0]={targ_src_fi.path}"
            )
            continue

        # Sanity checks:

        if is_file_info_list_bad_state(primary_hasher_name, template_fi_list):
            message = (
                f"ERROR: Template root file_info list bad state causes "
                f"target source file_info not to be arranged."
            )
            logging.error(message)
            raise InvalidStateError(message)

        if template_fi_list[0].primary_digest != targ_src_fi.primary_digest:
            message = (
                f"ERROR: Target source and template root digests do not match "
                f"when expected: {primary_hasher_name}: "
                f"target={targ_src_fi.primary_digest} "
                f"template={template_fi_list[0].primary_digest} "
                f"target_path={targ_src_fi.path} "
                f"template_path={template_fi_list[0].path}"
            )
            logging.error(message)
            raise InvalidStateError(message)

        template_match_count += 1

        logging.debug(
            f"Target source and template root digests match: "
            f"target={targ_src_fi.primary_digest} "
            f"template={template_fi_list[0].primary_digest} "
            f"target_path={targ_src_fi.path} "
            f"template_path={template_fi_list[0].path}"
        )

        template_fi_list.sort(key=lambda fi: fi.nc_path)

        # Determine relative path for target source.
        targ_src_rel_path = rel_path(
            top_level_dir=target_source_root,
            path=targ_src_fi.path,
        )
        targ_src_rel_path_nc = os.path.normcase(targ_src_rel_path)

        # Build relative path info list for all template candidates.
        template_rpi_list: list[_FileInfoRelPathInfo] = []
        for template_fi in template_fi_list:
            template_rpi_list.append(
                _FileInfoRelPathInfo(
                    fi=template_fi,
                    rel_path = rel_path(
                        top_level_dir=template_root,
                        path=template_fi.path,
                    ),
                )
            )

        # Find closest equal/greater lexicographic match.
        template_rpi_found = template_rpi_list[0]
        if len(template_rpi_list) > 1:
            for idx, template_rpi in enumerate(template_rpi_list):
                if targ_src_rel_path_nc >= template_rpi.rel_path_nc:
                    template_rpi_found = template_rpi
                    if targ_src_rel_path_nc > template_rpi.rel_path_nc and idx > 0:
                        template_rpi_found = template_rpi_list[idx - 1]
                    break

        # Mark target source for move to target dest using discovered template rel path.
        arrange_info = _ArrangeOperationInfo(
            template_fi=template_rpi_found.fi,
            target_source_fi=targ_src_fi,
            target_source_full_path=targ_src_fi.path,
            target_dest_rel_path=template_rpi_found.rel_path,
            target_dest_full_path=os.path.join(target_dest_root, template_rpi_found.rel_path),
        )
        arrange_oper_info_list.append(arrange_info)

        logging.debug(
            f"Arrange operation added to list: "
            f"{primary_hasher_name}: "
            f"{arrange_info.target_source_fi.primary_digest}: "
            f"{arrange_info.target_source_full_path} --> "
            f"{arrange_info.target_dest_full_path}"
        )

        if targ_src_fi.info_data_file_exists():
            # Add operation for sidecar .atbu file.
            template_info_data_rel_path = rel_path(
                top_level_dir=template_root,
                path=template_rpi_found.fi.info_data_file_path,
            )
            arrange_info = _ArrangeOperationInfo(
                template_fi=template_rpi_found.fi,
                target_source_fi=targ_src_fi,
                target_source_full_path=targ_src_fi.info_data_file_path,
                target_dest_rel_path=template_info_data_rel_path,
                target_dest_full_path=os.path.join(target_dest_root, template_info_data_rel_path),
            )
            arrange_oper_info_list.append(arrange_info)
            logging.debug(
                f"Arrange operation added to list: "
                f"{primary_hasher_name}: "
                f"{arrange_info.target_source_fi.primary_digest}: "
                f"{arrange_info.target_source_full_path} --> "
                f"{arrange_info.target_dest_full_path}"
            )

        # Remove selected template file info from candidates.
        template_fi_list.remove(template_rpi_found.fi)

    return (
        template_match_count,
        template_no_match_count,
        template_match_exhausted_count,
        arrange_oper_info_list
    )

def arrange_target(
    template_root: str,
    template_info: dict[str, list[FileInformationPersistent]],
    target_source_root: str,
    target_source_info: dict[str, list[FileInformationPersistent]],
    target_dest_root: str,
    undo_file_path: str,
    is_dryrun: bool,
):
    undofile = None
    arrange_undo_info: list[_ArrangeUndoInfo] = []
    successful_move_count = 0
    failed_move_count = 0
    target_dest_file_exists_count = 0
    target_source_file_count = sum(
        [(2 if fi.info_data_file_exists() else 1) for _, l in target_source_info.items() for fi in l]
    )
    try:
        # Ensure undo file can be created, readied for writing after completing arrange operations.
        if undo_file_path is not None:
            undofile = open(
                file=undo_file_path,
                 mode="w",
                 encoding="utf-8")
            logging.info(f"Undo file opened: {undo_file_path}")
        else:
            logging.info(f"Undo file will not be used.")

        (
            template_match_count,
            template_no_match_count,
            template_match_exhausted_count,
            arrange_oper_info_list,
        ) = _build_arrange_operation_list(
            template_root=template_root,
            template_info=template_info,
            target_source_root=target_source_root,
            target_source_info=target_source_info,
            target_dest_root=target_dest_root,
        )

        if len(arrange_oper_info_list) == 0:
            logging.info(
                f"No arrangement operations could be determined. Nothing to do."
            )
            return

        logging.info(
            f"Discovered {len(arrange_oper_info_list)} move target source to dest operations."
        )

        for arrange_oper_info in arrange_oper_info_list:
            try:
                if os.path.exists(arrange_oper_info.target_dest_full_path):
                    # While os.renames will throw the same exception on destination file already
                    # existing, it is exlicity thrown here to simply avoid the call to os.renames.
                    raise FileExistsError(
                        f"ERROR: The destination file already exists: "
                        f"{arrange_oper_info.target_dest_full_path}"
                    )
                if not is_dryrun:
                    os.renames(
                        old=arrange_oper_info.target_source_full_path,
                        new=arrange_oper_info.target_dest_full_path,
                    )
                successful_move_count += 1
                arrange_undo_info.append(
                    _ArrangeUndoInfo(
                        source_full_path=arrange_oper_info.target_source_fi.path,
                        dest_full_path=arrange_oper_info.target_dest_full_path,
                        is_move_successful=True,
                        error_msg=None if not is_dryrun else "<dryrun>",
                    )
                )
                logging.info(
                    f"Move successful{' (dry run)' if is_dryrun else ''}: "
                    f"{arrange_oper_info.target_source_fi.path}"
                    f" --> "
                    f"{arrange_oper_info.target_dest_full_path}"
                )
            except OSError as ex:
                failed_move_count += 1
                if isinstance(ex, FileExistsError):
                    target_dest_file_exists_count += 1
                error_message = (
                    f"Move failed{' (dry run)' if is_dryrun else ''}: "
                    f"{arrange_oper_info.target_source_fi.path}"
                    f" --> "
                    f"{arrange_oper_info.target_dest_full_path} "
                    f"{exc_to_string(ex)}"
                )
                logging.error(error_message)
                arrange_undo_info.append(
                    _ArrangeUndoInfo(
                        source_full_path=arrange_oper_info.target_source_fi.path,
                        dest_full_path=arrange_oper_info.target_dest_full_path,
                        is_move_successful=False,
                        error_msg=error_message,
                    )
                )
    except Exception as ex:
        arrange_undo_info.append(
            _ArrangeUndoInfo(
                source_full_path="<uncaught_exception>",
                dest_full_path="<uncaught_exception>",
                is_move_successful=False,
                error_msg=(
                    f"Arrange processing ended abruptly, undo information may be incomplete "
                    f"{' (dry run)' if is_dryrun else ''}: "
                    f"{exc_to_string(ex)}"
                )
            )
        )
        raise
    finally:
        if undofile is not None:
            logging.info(
                f"Writing"
                f"{' (dry run) ' if is_dryrun else ' '}"
                f"undo information to undo file '{undo_file_path}' ...")
            undofile.write(
                json.dumps(
                    obj=arrange_undo_info,
                    indent=4,
                    cls=_ArrangeUndoInfo.get_json_encoder(),
                )
            )
            undofile.close()

    dryrun_str = "(--whatif) " if is_dryrun else " "

    logging.info(
        f"{'Total target source files ' + dryrun_str:.<65} "
        f"{target_source_file_count} (includes any sidecar {ATBU_PERSISTENT_INFO_EXTENSION} files)"
    )
    logging.info(
        f"{'Total arrange operations ' + dryrun_str:.<65} "
        f"{len(arrange_oper_info_list)} (includes any sidecar {ATBU_PERSISTENT_INFO_EXTENSION} files)"
    )
    logging.info(
        f"{'Successful move operations ' + dryrun_str:.<65} "
        f"{successful_move_count} (includes any sidecar {ATBU_PERSISTENT_INFO_EXTENSION} files)"
    )
    logging.info(
        f"{'Failed move operations ' + dryrun_str:.<65} "
        f"{failed_move_count}"
    )
    logging.info(
        f"{'Failures due to destination already exists ' + dryrun_str:.<65} "
        f"{target_dest_file_exists_count}"
    )
    logging.info(
        f"{'Total target source files not moved ' + dryrun_str:.<65} "
        f"{target_source_file_count - successful_move_count} (includes any sidecar {ATBU_PERSISTENT_INFO_EXTENSION} files)"
    )
    logging.info(
        f"{'Total template digest matches ' + dryrun_str:.<65} "
        f"{template_match_count} (1 count for each file and accompanying sidecar file)"
    )
    logging.info(
        f"{'Total template digest matches not found ' + dryrun_str:.<65} "
        f"{template_no_match_count}"
    )
    logging.info(
        f"{'Total template digest matches exhausted ' + dryrun_str:.<65} "
        f"{template_match_exhausted_count}"
    )
    logging.info(
        f"{'Undo file ' + dryrun_str:.<65} "
        f"{'<none>' if undo_file_path is None else undo_file_path}"
    )

    logging.info(
        f"Arrange complete."
    )

def handle_arrange(args):
    if not args.no_undo and args.undofile is None:
        raise InvalidCommandLineArgument(
            f"Either --no-undofile or --undofile <path> must be specified."
        )
    locations = extract_location_info(
        args.locations,
        min_required=3,
        max_allowed=3,
        must_exist=False
    )
    loc_template_root = locations[0][0]
    loc_template_root_per_type = locations[0][1]
    if not os.path.exists(loc_template_root):
        raise InvalidCommandLineArgument(
            f"The template root path does not exist: {loc_template_root}"
        )

    loc_target_source_root = locations[1][0]
    loc_target_source_root_per_type = locations[1][1]
    if not os.path.exists(loc_target_source_root):
        raise InvalidCommandLineArgument(
            f"The target source root does not exist: {loc_target_source_root}"
        )

    loc_target_dest_root = locations[2][0]
    loc_target_dest_root_per_type = locations[2][1]

    sr_target_source = os.stat(loc_target_source_root)
    if not os.path.exists(loc_target_dest_root):
        os.makedirs(loc_target_dest_root, exist_ok=True)
    sr_target_dest = os.stat(loc_target_dest_root)
    if sr_target_source.st_dev != sr_target_dest.st_dev:
        raise ValueError(
            f"The target source root and destination must be located on the same drive. "
            f"target source "
            f"st_dev=0x{hex(sr_target_source.st_dev)} "
            f"dest st_dev=0x{hex(sr_target_dest.st_dev)}"
        )

    loc_template_DBs = FileInformationDatabaseCollection(
        source_path=loc_template_root, persist_types=loc_template_root_per_type
    )

    loc_target_source_DBs = FileInformationDatabaseCollection(
        source_path=loc_target_source_root, persist_types=loc_target_source_root_per_type
    )

    logging.info(f"{'Template root directory ':.<45} {loc_template_root}")
    logging.info(f"{'Template root persist type ':.<45} {loc_template_root_per_type}")
    logging.info(f"{'Target source root directory ':.<45} {loc_target_source_root}")
    logging.info(f"{'Target source root persist type ':.<45} {loc_target_source_root_per_type}")
    logging.info(f"{'Target source destination directory ':.<45} {loc_target_dest_root}")
    logging.info(f"{'Target source destination persist type ':.<45} {loc_target_dest_root_per_type}")

    logging.info(f"-" * 65)
    logging.info(f"Searching template root directory: {loc_template_root}")
    updaterA = loc_template_DBs.update(
        change_detection_type=args.change_detection_type,
        update_stale=args.update_stale,
        whatif=args.whatif,
    )
    loc_template_info = loc_template_DBs.get_dict_digest_to_fi()
    loc_template_info_skipped = updaterA.skipped_files

    logging.info(f"-" * 65)
    logging.info(f"Searching target source directory: {loc_target_source_root}")
    updaterB = loc_target_source_DBs.update(
        change_detection_type=args.change_detection_type,
        update_stale=args.update_stale,
        whatif=args.whatif,
    )
    loc_target_source_info = loc_target_source_DBs.get_dict_digest_to_fi()
    loc_target_source_skipped = updaterB.skipped_files

    arrange_target(
        template_root=loc_template_DBs.source_path,
        template_info=loc_template_info,
        target_source_root=loc_target_source_DBs.source_path,
        target_source_info=loc_target_source_info,
        target_dest_root=loc_target_dest_root,
        undo_file_path=args.undofile,
        is_dryrun=args.whatif,
    )
